"""Use a classifier as a regressor.

By default return predicted probabilities as numeric predictions.

"""


from sklearn.base import BaseEstimator, RegressorMixin

class ClassifierAsRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, estimator, response_method='predict_proba',
                 pos_label=1):
        """Wraps a classifier such that it behaves like a regressor.

        The predict method return by default predicted probability for
        class specified by pos_label (default 1).  The method used for
        prediction can be changed by passing the response_method
        argument.

        If response_method return a vector (e.g. decision_function)
        pos_label will be ignored.

        """
        self.estimator = estimator
        self.response_method = response_method
        self.pos_label = pos_label
    def fit(self, *args, **kwargs):
        self.fitted_estimator_ = self.estimator.fit(*args, **kwargs)
        return self
    def predict(self, *args, **kwargs):
        resp_method = getattr(self.fitted_estimator_, self.response_method)
        preds = resp_method(*args, **kwargs)
        pred_ndim = len(preds.shape)
        if pred_ndim > 2:
            raise RuntimeError("ClassifierAsRegressor: response method"
                               " must return a vector of a matrix.")
        elif pred_ndim == 2:
            preds = preds[:,self.pos_label]
        return preds

    def __getattr__(self, name):
        if name in ["fitted_estimator_", "response_method", "pos_label"]:
            try:
                return self.__dict__[name]
            except:
                raise AttributeError(f"ClassifierAsRegressor has"
                                     " no attribute {name}")
        if "fitted_estimator_" not in self.__dict__:
            return getattr(self.estimator, name)
        return getattr(self.fitted_estimator_, name)



"""The colon datasets from R survival package.

"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="colon.csv",
    url=('https://vincentarelbundock.github.io/'
         'Rdatasets/csv/survival/colon.csv'),
    checksum=('6f3472a64f696e3195daa198f054180c'
              '3e4c66408f7fb8c548c6f4c7b8f898ee'))

def _float_w_nan(x):
    """Convert strings to floats with empty strings converted to
    nan's."""
    y = [v if v != "" else "nan" for v in x]
    return np.array(y, float), float

def fetch_colon(data_home=None, download_if_missing=True,
                random_state=None, shuffle=False,
                categ_as_strings=False, return_X_y=False,
                as_frame=False):
    """Load the colon dataset from R survival package (uplift survival).

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_recurrence_time : numpy array
        Survival or censoring time.

    dataset.target_recurrence_status : numpy array
        Censoring. Each value is 0 (censored) or 1 (event).

    dataset.target_death_time : numpy array
        Survival or censoring time.

    dataset.target_death_status : numpy array
        Censoring. Each value is 0 (censored) or 1 (event).

    dataset.DESCR : string
        Description of the dataset.

    (data, target_time, target_status) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['Obs', 'Lev', 'Lev+5FU']
    differ_values = {"1":"well", "2":"moderate", "3":"poor", "":"NA"}     
    extent_values = {"1":"submucosa", "2":"muscle", "3":"serosa", 
                     "4":"contiguous_structures"} 

    # attribute descriptions
    treatment_descr = [("rx", treatment_values)]
    target_descr = [("target_recurrence_time", float, "time"),
                    ("target_recurrence_status", np.int32, "status"),]

    feature_descr = [#("rownames", np.int32),
                     #("id", np.int32),
                     #("study", np.int32),
                     ("sex", np.int32),
                     ("age", float),
                     ("obstruct", np.int32),
                     ("perfor", np.int32),
                     ("adhere", np.int32),
                     ("nodes", _float_w_nan),
                     ("differ", differ_values),
                     ("extent", extent_values),
                     ("surg", np.int32),
                     ("node4", np.int32),
                     ("etype", np.int32),
    ]

    ret = _fetch_remote_csv(ARCHIVE, "colon",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=17
                            )
    if not return_X_y:
        ret.descr = __doc__

    # extract different targets
    ret.target_names = ["target_recurrence_time", "target_recurrence_status",
                        "target_death_time", "target_death_status",]
    ret.feature_names.remove("etype")
    if as_frame:
        etype = ret.data["etype"]
        ret.data = ret.data[etype==1].reset_index().drop("etype", axis=1)
    else:
        etype = ret.data[:, -1]
        ret.data = ret.data[etype==1,:-1]
    ret.treatment = ret.treatment[etype==1]
    ret.target_death_time = ret.target_recurrence_time[etype==2]
    ret.target_death_status = ret.target_recurrence_status[etype==2]
    ret.target_recurrence_time = ret.target_recurrence_time[etype==1]
    ret.target_recurrence_status = ret.target_recurrence_status[etype==1]
    return ret



"""The Starbucks dataset.

"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="training.csv",
    url=('https://raw.githubusercontent.com/01KAT1/'
         'Marketing-Promotion-Campaign-Uplift-Modelling-Starbucks-Dataset/'
         'main/training.csv'),
    checksum=('2ac9d5c601b134b9b69e742b97d01652'
              '9c8082f6a611d5b243ab0c5b2ceaf83e'))

def fetch_Starbucks(data_home=None, download_if_missing=True,
                    random_state=None, shuffle=False,
                    categ_as_strings=False, return_X_y=False,
                    as_frame=False):
    """Load the Starbucks dataset.

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_purchase : numpy array
        Indicator whether a purchase was made.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_purchase) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['No', 'Yes']
    V1_values = ["0", "1", "2", "3"]
    V4_values = ["1", "2"]
    V5_values = ["1", "2", "3", "4"]
    V6_values = ["1", "2", "3", "4"]
    V7_values = ["1", "2"]

    # attribute descriptions
    treatment_descr = [("Promotion", treatment_values)]
    target_descr = [("target_purchase", np.int32, "purchase"),]

    feature_descr = [#("id", np.int32),
                     ("V1", V1_values),
                     ("V2", float),
                     ("V3", float),
                     ("V4", V4_values),
                     ("V5", V5_values),
                     ("V6", V6_values),
                     ("V7", V7_values),
    ]

    ret = _fetch_remote_csv(ARCHIVE, "Starbucks",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=10
                            )
    if not return_X_y:
        ret.descr = __doc__

    return ret



"""Uplift models based on multiple classification/regression models."""

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model._base import LinearModel

from .base import UpliftMetaModelBase
from ..base import UpliftRegressorMixin
from ..base import UpliftClassifierMixin



class _MultimodelUpliftModel(UpliftMetaModelBase):
    """Base class fot multimodels (T-learners).

    if ignore_control is set to True, response model on treatment is
    built.

    """
    def __init__(self, base_estimator, prediction_method,
                 ignore_control=False):
        super().__init__(base_estimator)
        self.prediction_method = prediction_method
        self.ignore_control = ignore_control
    def _get_model_names_list(self, X=None, y=None, trt=None):
        control_m_name = "_model_c" if self.ignore_control else "model_c"
        m_names = []
        for i in range(self.n_trt_ + 1):
            if i == 0:
                name = control_m_name
            else:
                name = "model_t"
                if self.n_trt_ > 1:
                    name += str(i-1)
            m_names.append(name)
        return m_names
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        """Return training sets for all models in the meta model.

        Each iteration returns a triple of predictor matrix, target
        vector, and sample weights (possibly None).

        """
        for i in range(self.n_models_):
            if self.ignore_control and i == 0:
                yield None, None, None
            else:
                mask = (trt==i)
                if sample_weight is None:
                    wi = None
                else:
                    wi = sample_weight[mask]
                Xi = X[mask]
                yi = y[mask]
                yield Xi, yi, wi
    def _predict_diffs(self, X):
        """Predict differences between model predictions for each
        treatment."""
        if self.ignore_control:
            y_control = 0
        else:
            y_control = getattr(self.models_[0][1], self.prediction_method)(X)
        pred_diffs = [getattr(self.models_[i+1][1], self.prediction_method)(X)
                      - y_control for i in range(self.n_trt_)]
        return pred_diffs


class MultimodelUpliftRegressor(UpliftRegressorMixin, _MultimodelUpliftModel):
    """Multimodel uplift regressor.

    Build separate models for control and all treatments, subtract
    control predictions from treatment predictions.

    Parameters
    ----------

    base_estimator : a sklearn regressor or list of (name, regressor)
        tuples.  If a list is provided the first model is used for
        control, successive one for treatments.  If different parameters are
        to be used per each model, the list version must be given.  If a
        single estimator is given it will be cloned for every treatment.
    """
    def __init__(self, base_estimator=LinearRegression(), ignore_control=False):
        super().__init__(base_estimator, "predict",
                         ignore_control=ignore_control)
    def predict(self, X):
        pred_diffs = self._predict_diffs(X)
        if self.n_trt_ == 1:
            y = pred_diffs[0]
        else:
            y = np.column_stack(pred_diffs)
        return y

class _MultimodelUpliftClassifierBase(UpliftClassifierMixin, _MultimodelUpliftModel):
    def __init__(self, base_estimator, ignore_control=False):
        super().__init__(base_estimator, prediction_method="predict_proba",
                         ignore_control=ignore_control)
    def predict(self, X):
        pred_diffs = self._predict_diffs(X)
        if self.n_trt_ == 1:
            y = pred_diffs[0]
        else:
            y = np.dstack(pred_diffs)
        return y
    """Abstract base for several multimodel and response uplift
    classifiers.

    Implements the predict method."""

class MultimodelUpliftClassifier(_MultimodelUpliftClassifierBase):
    """Multimodel uplift classifier.

    Build separate models for control and all treatments, subtract
    control predicted probs from treatment predicted probs.

    Parameters
    ----------

    base_estimator : a sklearn classifier supporting predict_proba or
        list of (name, regressor) tuples.  If a list is provided the
        first model is used for control, successive one for
        treatments.  If different parameters are to be used per each
        model, the list version must be given.  If a single estimator
        is given it will be cloned for every treatment.

    """
    def __init__(self, base_estimator=LogisticRegression()):
        super().__init__(base_estimator)

class MultimodelUpliftLinearRegressor(MultimodelUpliftRegressor, LinearModel):
    """Uplift regressor with coef_ and intercept_ fields."""
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ui = self.models_[i+1][1].coef_ - c0
            ii = self.models_[i+1][1].intercept_ - i0
            coef[i,:] = ui
            intercept[i] = ii
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)

class MultimodelUpliftLinearRegressorJamesSeparate(MultimodelUpliftRegressor, LinearModel):
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_
        coef_all0 = np.concatenate((np.array([i0]),c0))
        alpha0 = (1-(self.p-3)*self.sigma[0]**2/((coef_all0-np.mean(coef_all0))@self.X_2[0]@(coef_all0-np.mean(coef_all0))))
        
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ci = self.models_[i+1][1].coef_
            ii2 = self.models_[i+1][1].intercept_
            coef_alli = np.concatenate((np.array([ii2]),ci))
            alphai = (1-(self.p-3)*self.sigma[i+1]**2/((coef_alli-np.mean(coef_alli))@self.X_2[i+1]@(coef_alli-np.mean(coef_alli))))
            ui = alphai*ci - alpha0*c0
            #print(c0)
            ii = alphai*ii2 - alpha0*i0
            coef[i,:] = ui
            intercept[i] = ii
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)

class MultimodelUpliftLinearRegressorJamesU(MultimodelUpliftRegressor, LinearModel):
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_

        
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ci = self.models_[i+1][1].coef_
            ii2 = self.models_[i+1][1].intercept_
            ui = ci - c0
            #print(c0)
            ii = ii2 - i0
            coef_all = np.concatenate((np.array([ii]),ui))
            alpha = (1-(self.p-3)/((coef_all -np.mean(coef_all))@np.linalg.pinv(self.sigma[i+1]**2*np.linalg.pinv(self.X_2[i+1])+self.sigma[0]**2*np.linalg.pinv(self.X_2[0]))@(coef_all-np.mean(coef_all))))        
            coef[i,:] = alpha*(ui-np.mean(ui))+np.mean(ui)
            intercept[i] = alpha*(ii-np.mean(ii))+np.mean(ii)
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)


class MultimodelUpliftLinearRegressorMSESeparate(MultimodelUpliftRegressor, LinearModel):
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_
        var_beta0 = self.sigma[0]**2*np.linalg.pinv(self.X_2[0])
        coef_all0 = np.concatenate((np.array([i0]),c0))
        W0 = self.X_2[0]/self.X_2[0].shape[0]
        alpha0 = coef_all0.T@W0@coef_all0/(coef_all0.T@W0@coef_all0+np.trace(W0@var_beta0))
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ci = self.models_[i+1][1].coef_
            ii2 = self.models_[i+1][1].intercept_
            var_betai = self.sigma[i+1]**2*np.linalg.pinv(self.X_2[i+1])
            coef_alli = np.concatenate((np.array([ii2]),ci))
            Wi = self.X_2[i+1]/self.X_2[i+1].shape[0]
            alphai = coef_alli.T@Wi@coef_alli/(coef_alli.T@Wi@coef_alli+np.trace(Wi@var_betai))
            ui = alphai*ci - alpha0*c0
            #print(c0)
            ii = alphai*ii2 - alpha0*i0
            coef[i,:] = ui
            intercept[i] = ii
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)
    
    
class MultimodelUpliftLinearRegressorMSEU(MultimodelUpliftRegressor, LinearModel):
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_
        var_beta0 = self.sigma[0]**2*np.linalg.pinv(self.X_2[0])
        coef_all0 = np.concatenate((np.array([i0]),c0))
        
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ci = self.models_[i+1][1].coef_
            ii2 = self.models_[i+1][1].intercept_
            var_betai = self.sigma[i+1]**2*np.linalg.pinv(self.X_2[i+1])
            coef_alli = np.concatenate((np.array([ii2]),ci))
            ui = ci - c0
            #print(c0)
            ii = ii2 - i0
            coef_all = np.concatenate((np.array([ii]),ui))
            W = (self.X_2[0]+self.X_2[i+1])/(self.n_[0]+self.n_[i+1])
            A = np.column_stack((np.vstack((coef_alli.T@W@coef_alli+np.trace(W@var_betai),-coef_all0.T@W@coef_alli)),
                         np.vstack((-coef_alli.T@W@coef_all0,coef_all0.T@W@coef_all0+np.trace(W@var_beta0)))))
    
            b = np.array([coef_alli.T@W@coef_all,-coef_all0.T@W@coef_all]) 
            alphas = np.linalg.pinv(A)@b
    
            alpha_t = alphas[0]
            alpha_c = alphas[1]
            
            coef[i,:] = alpha_t*ci - alpha_c*c0
            intercept[i] = alpha_t*ii2 - alpha_c*i0
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)    
    



"""
The :mod:`usklearn.utils` module includes various utilities.
"""

from .validation import check_trt
from .multi_array import MultiArray
from .array_utils import safe_hstack
from .metrics import area_under_curve

__all__ = ["check_trt", "MultiArray", "safe_hstack", "area_under_curve"]



from sklearn.linear_model import LinearRegression

from usklearn.classifiers import MemoizedClassifier

def test_memoized_classifier(tmp_path):
    cache = tmp_path / "cache"
    lr = LinearRegression()
    mlr = MemoizedClassifier(lr, str(cache))
    X = [[1.0,0], [2,1], [2,3]]
    y = [1.0, 2.0, 3.0]
    mlr.fit(X, y)
    
    lr2 = LinearRegression()
    mlr2 = MemoizedClassifier(lr, str(cache))
    mlr2.fit(X, y)



"""Utility functions used for metric computation."""

import numpy as np

def area_under_curve(xs, ys, subtract_diag=True):
    """Compute area under a curve given by xs and ys.

    If subtract_diag is True area under the diagonal is subtracted.

    """
    auc = np.trapz(ys, xs)
    if subtract_diag:
        a = xs[-1] - xs[0]
        h = ys[-1]
        auc -= a*h/2
    return auc



"""The Megafon dataset.

The version used comes from the scikit-uplift package.
"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="megafon_dataset.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Megafon/megafon_dataset.csv.gz'),
    checksum=('cdcb2d052b90f8eefa75937d5540f114'
              'd0748ea231d95e2778dd6760478e4a00'))

def fetch_Megafon(data_home=None, download_if_missing=True,
                    random_state=None, shuffle=False,
                    categ_as_strings=False, return_X_y=False,
                    as_frame=False):
    """Load the Megafon dataset.

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_conversion : numpy array
        Indicator whether a conversion occurred.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_conversion) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['control', 'treatment']

    # attribute descriptions
    treatment_descr = [("treatment_group", treatment_values)]
    target_descr = [("target_conversion", np.int32, "conversion"),]

    feature_descr = [("X_"+str(i+1), float) for i in range(50)]

    ret = _fetch_remote_csv(ARCHIVE, "Megafon",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=52
                            )
    if not return_X_y:
        ret.descr = __doc__

    return ret



"""Utilities for input validation."""

import numpy as np

from sklearn.utils import column_or_1d, check_consistent_length

def check_trt(trt, n_trt=None, y=None):
    trt = column_or_1d(trt)
    if not np.issubdtype(trt.dtype, np.integer):
        raise ValueError("Treatment values must be integers")
    if (trt < 0).any():
        raise ValueError("Treatment values must be >= 0")
    if n_trt is not None:
        if np.max(trt) > n_trt:
            raise ValueError("Treatment values must be <= n_trt")
    else:
        n_trt = np.max(trt)
    return trt, n_trt




from usklearn.metrics import uplift_curve

import numpy as np

def test_small_uplift_curve():
    y_true = [0,1,0,0,0,1,1]
    score  = [1,2,3,1,2,2,3]
    trt    = [0,0,0,1,1,1,1]
    x, u = uplift_curve(y_true, score, trt)
    assert x[0] == 0
    assert u[0] == 0
    assert x[-1] == 1
    assert np.allclose(u[-1], 1/2 - 1/3)



"""Model validation functions for uplift models."""


import numpy as np

from sklearn.base import is_classifier, BaseEstimator
from sklearn.base import MetaEstimatorMixin
from sklearn.utils import indexable
from sklearn.utils.metaestimators import available_if
from sklearn.model_selection import check_cv
from sklearn.model_selection import cross_validate as _sklearn_cross_validate
from sklearn.model_selection import cross_val_predict as _sklearn_cross_val_predict
from sklearn.model_selection import permutation_test_score as _sklearn_permutation_test_score
from sklearn.model_selection import learning_curve as _sklearn_learning_curve
from sklearn.preprocessing import LabelEncoder

from sklearn.utils.metaestimators import _BaseComposition

from ..utils import check_trt
from ..utils import MultiArray

from ..metrics import check_uplift_scoring


__all__ = ['cross_validate', 'cross_val_score', 'cross_val_predict',
           'permutation_test_score', 'learning_curve', 'validation_curve']

# Adapted from sklearn:
def _estimator_has(attr):
    """Check if we can delegate a method to the underlying estimator.
    """

    def check(self):
        # raise an AttributeError if `attr` does not exist
        getattr(self.base_estimator, attr)
        return True

    return check
def _estimator_has_predict():
    """Check the predict method of wrapped uplift estimator.
    """

    def check(self):
        # raise an AttributeError if `attr` does not exist
        if self.wrapped_predict is None:
            getattr(self.base_estimator, "predict")
        else:
            getattr(self.base_estimator, self.wrapped_predict)
        return True

    return check

def _extract_uplift_arrays(X):
        """Extract data for uplift training from MultiArray."""
        real_X = X.main_array
        y = X.array_dict["y"]
        trt = X.array_dict["trt"]
        n_trt = X.scalar_dict["n_trt"]
        return real_X, y, trt, n_trt
def _extract_main_array(X):
        """Extract main array if input is a MultiArray, otherwise, just return X.

        Useful in X is used for prediction on a fitted model."""
        if isinstance(X, MultiArray):
            X = X.main_array
        return X

class _WrappedUpliftEstimator(_BaseComposition, MetaEstimatorMixin, BaseEstimator):
    """Wrap upift estimator inside a sklearn estimator interface.

    If wrapped_predict is set it specifies the method to be called by
    wrapped model's predict.  This way standard scikit functions can
    access uplift specific predictions such as predict_action.
    """
    _uplift_model = True
    def __init__(self, base_estimator, wrapped_predict=None):
        self.base_estimator = base_estimator
        self.wrapped_predict = wrapped_predict
        if (self.wrapped_predict is not None and not
            hasattr(self.base_estimator, wrapped_predict)):
            raise RuntimeError(f"Can't wrap uplift model: {self.wrapped_predict} method not available.")
    @property
    def _estimator_type(self):
        return self.base_estimator._estimator_type
    def fit(self, X, y, **kwargs):
        real_X, y, trt, n_trt = _extract_uplift_arrays(X)
        return self.base_estimator.fit(real_X, y, trt, n_trt, **kwargs)
    def score(self, X, y, *args, **kwargs):
        real_X, y, trt, n_trt = _extract_uplift_arrays(X)
        return self.base_estimator.score(real_X, y, trt, n_trt, *args, **kwargs)
    @available_if(_estimator_has_predict())
    def predict(self, X):
        X = _extract_main_array(X)
        if self.wrapped_predict is not None:
            return getattr(self.base_estimator, self.wrapped_predict)(X)
        return self.base_estimator.predict(X)
    @available_if(_estimator_has("predict_action"))
    def predict_action(self, X):
        X = _extract_main_array(X)
        return self.base_estimator.predict_action(X)
    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X):
        X = _extract_main_array(X)
        return self.base_estimator.predict_proba(X)
    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X):
        X = _extract_main_array(X)
        return self.base_estimator.predict_log_proba(X)
    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        X = _extract_main_array(X)
        return self.base_estimator.decision_function(X)
    @available_if(_estimator_has("transform"))
    def transform(self, X):
        X = _extract_main_array(X)
        return self.base_estimator.transform(X)
    @available_if(_estimator_has("inverse_transform"))
    def inverse_transform(self, Xt):
        X = _extract_main_array(X)
        return self.base_estimator.inverse_transform(Xt)
    def get_params(self, deep=True):
        if hasattr(self.base_estimator, "fit"):
            return super().get_params(deep=deep)
        return self._get_params('base_estimator', deep=deep)
    def set_params(self, **kwargs):
        if hasattr(self.base_estimator, "fit"):
            return super().set_params(**kwargs)
        else:
            self._set_params('base_estimator', **kwargs)
        return self
    def get_metadata_routing(self):
        return self.base_estimator.get_metadata_routing()
    def __getattr__(self, arg):
        """Delegate to real grid search."""
        return getattr(self.base_estimator, arg)

class _WrappedScoring:
    """Wrap uplift scoring into a sklearn scoring interface."""
    def __init__(self, uplift_scoring):
        self.uplift_scoring = uplift_scoring
    def __call__(self, estimator, X, y, *args, **kwargs):
        real_X, y, trt, n_trt = _extract_uplift_arrays(X)
        return self.uplift_scoring(estimator.base_estimator, real_X, y, trt, n_trt, *args, **kwargs)

# copied from sklearn to avoid use of private sklearn functions
def _check_multimetric_scoring(estimator, scoring):
    err_msg_generic = (
        f"scoring is invalid (got {scoring!r}). Refer to the "
        "scoring glossary for details: "
        "https://scikit-learn.org/stable/glossary.html#term-scoring"
    )

    if isinstance(scoring, (list, tuple, set)):
        err_msg = (
            "The list/tuple elements must be unique strings of predefined scorers. "
        )
        try:
            keys = set(scoring)
        except TypeError as e:
            raise ValueError(err_msg) from e

        if len(keys) != len(scoring):
            raise ValueError(
                f"{err_msg} Duplicate elements were found in"
                f" the given list. {scoring!r}"
            )
        elif len(keys) > 0:
            if not all(isinstance(k, str) for k in keys):
                if any(callable(k) for k in keys):
                    raise ValueError(
                        f"{err_msg} One or more of the elements "
                        "were callables. Use a dict of score "
                        "name mapped to the scorer callable. "
                        f"Got {scoring!r}"
                    )
                else:
                    raise ValueError(
                        f"{err_msg} Non-string types were found "
                        f"in the given list. Got {scoring!r}"
                    )
            scorers = {
                scorer: check_uplift_scoring(estimator, scoring=scorer) for scorer in scoring
            }
        else:
            raise ValueError(f"{err_msg} Empty list was given. {scoring!r}")

    elif isinstance(scoring, dict):
        keys = set(scoring)
        if not all(isinstance(k, str) for k in keys):
            raise ValueError(
                "Non-string types were found in the keys of "
                f"the given dict. scoring={scoring!r}"
            )
        if len(keys) == 0:
            raise ValueError(f"An empty dict was passed. {scoring!r}")
        scorers = {
            key: check_uplift_scoring(estimator, scoring=scorer)
            for key, scorer in scoring.items()
        }
    else:
        raise ValueError(err_msg_generic)
    return scorers


def uplift_check_cv(cv, y, trt, n_trt, *, classifier=False):
    """Return a correct cv and y_stratify to pass to

    _sklearn_cross_validate.

    y_stratify is used only for stratification since y is contained in a
    MultiArray."""
    
    # always stratify on treatment and, if available, also on class
    if classifier:
        le = LabelEncoder()
        y_stratify = le.fit_transform(y)
        y_stratify = y_stratify * (n_trt+1) + trt
    else:
        y_stratify = trt
    # classifier=True ensures stratification
    cv = check_cv(cv, y_stratify, classifier=True)
    return cv, y_stratify

        
def cross_validate(estimator, X, y, trt, n_trt=None, groups=None, scoring=None, cv=None,
                       *args, **kwargs):
    X, y, trt, groups = indexable(X, y, trt, groups)
    trt, n_trt = check_trt(trt, n_trt)
    # y_stratify is used only for stratification
    cv, y_stratify = uplift_check_cv(cv, y, trt, n_trt, classifier=is_classifier(estimator))
    # multiarray to pass additional data
    Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
    # wrapped estimator
    wrapped_est = _WrappedUpliftEstimator(estimator)
    # wrapped scoring
    # check_cv explicitly to stratify on treatment even for regression
    scorers = _check_multimetric_scoring(estimator, scoring=scoring)
    wrapped_scorers = {name:_WrappedScoring(scoring) for name, scoring in scorers.items()}
    return _sklearn_cross_validate(wrapped_est, Xm, y_stratify, groups=groups, scoring=wrapped_scorers,
                                   cv=cv, *args, **kwargs)




def cross_val_score(estimator, X, y, trt, n_trt=None, groups=None, scoring=None, cv=None,
                    *args, **kwargs):
    # To ensure multimetric format is not supported
    scorer = check_uplift_scoring(estimator, scoring=scoring)
    cv_results = cross_validate(estimator=estimator, X=X, y=y, trt=trt,
                                n_trt=n_trt, groups=groups,
                                scoring={'score': scorer}, cv=cv,
                                *args, **kwargs)
    return cv_results['test_score']


def cross_val_predict(estimator, X, y, trt, n_trt=None, groups=None,
                      cv=None, method='predict', *args, **kwargs):
    X, y, trt, groups = indexable(X, y, trt, groups)
    trt, n_trt = check_trt(trt, n_trt)
    # y_stratify is used only for stratification
    cv, y_stratify = uplift_check_cv(cv, y, trt, n_trt, classifier=is_classifier(estimator))
    # multiarray to pass additional data
    Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
    # wrapped estimator
    if method not in ["predict", "predict_proba", "predict_log_proba",
                      "decision_function", "transform", "inverse_transform"]:
        wrapped_predict = method
        method = "predict"
    else:
        wrapped_predict = None
    wrapped_est = _WrappedUpliftEstimator(estimator, wrapped_predict=wrapped_predict)
    return _sklearn_cross_val_predict(wrapped_est, Xm, y_stratify, groups=groups, cv=cv,
                                      method=method, *args, **kwargs)


# permutation_test_score needs a modified approach to wrapping after
# wrapping y contains only stratification information which is
# permuted by this function

from sklearn.model_selection import BaseCrossValidator
class _FixedStrataCV(BaseCrossValidator):
    """Wrap a crossvalidator such that statification variable is the same
    for each iteration.
    """
    def __init__(self, cv, strata):
        self.cv = cv
        self.strata = strata
    def get_n_splits(self, X, y=None, groups=None):
        return self.cv.get_n_splits(X=X, y=y, groups=groups)
    def _iter_test_masks(self, X=None, y=None, groups=None):
        return self.cv._iter_test_masks(X=X, y=self.strata, groups=groups)
    def _iter_test_indices(self, X=None, y=None, groups=None):
        return self.cv._iter_test_indices(X=X, y=self.strata, groups=groups)

class _WrappedUpliftEstimatorOrigY(_WrappedUpliftEstimator):
    """Modification of wrapped uplift estimator which uses original y for
    training.
    """
    def fit(self, X, y, **kwargs):
        real_X, y_wrapped, trt, n_trt = _extract_uplift_arrays(X)
        return self.base_estimator.fit(real_X, y, trt, n_trt, **kwargs)
    def score(self, X, y, *args, **kwargs):
        real_X, y_wrapped, trt, n_trt = _extract_uplift_arrays(X)
        return self.base_estimator.score(real_X, y, trt, n_trt, *args, **kwargs)

    
def permutation_test_score(estimator, X, y, trt, n_trt=None, stratify_on_trt=True,
                           groups=None, cv=None,
                           scoring=None, *args, **kwargs):
    """Permutation score test for uplift models.

    If stratify_on_trt is true, permutations are performed separately
    for each treatment.  This way constant treatment effect will not
    affect the results.
    """
    X, y, trt, groups = indexable(X, y, trt, groups)
    trt, n_trt = check_trt(trt, n_trt)
    # y_stratify is used only for stratification
    cv, y_stratify = uplift_check_cv(cv, y, trt, n_trt, classifier=is_classifier(estimator))
    wrapped_cv = _FixedStrataCV(cv, y_stratify) # don't permute strata
    if stratify_on_trt:
        assert groups is None
        groups = trt
    Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
    wrapped_est = _WrappedUpliftEstimatorOrigY(estimator)
    scorer = check_uplift_scoring(estimator, scoring=scoring)
    wrapped_scorer = _WrappedScoring(scorer)
    return _sklearn_permutation_test_score(wrapped_est, Xm, y, groups=groups,
                                           cv=wrapped_cv, scoring=wrapped_scorer,
                                           *args, **kwargs)



def learning_curve(estimator, X, y, trt, n_trt=None, groups=None,
                   cv=None, scoring=None,
                   *args, **kwargs):
    X, y, trt, groups = indexable(X, y, trt, groups)
    trt, n_trt = check_trt(trt, n_trt)
    # y_stratify is used only for stratification
    cv, y_stratify = uplift_check_cv(cv, y, trt, n_trt, classifier=is_classifier(estimator))
    # multiarray to pass additional data
    Xm = MultiArray(X, array_dict={"y":y, "trt":trt}, scalar_dict={"n_trt":n_trt})
    wrapped_est = _WrappedUpliftEstimator(estimator)
    scorer = check_uplift_scoring(estimator, scoring=scoring)
    wrapped_scorer = _WrappedScoring(scorer)
    return _sklearn_learning_curve(wrapped_est, Xm, y_stratify, groups=groups, cv=cv,
                                   scoring = scorer, *args, **kwargs)




def validation_curve(estimator, X, y, param_name, param_range, groups=None,
                     cv='warn', scoring=None, n_jobs=None, pre_dispatch="all",
                     verbose=0, error_score='raise-deprecating'):
    """Validation curve.

    Determine training and test scores for varying parameter values.

    Compute scores for an estimator with different values of a specified
    parameter. This is similar to grid search with one parameter. However, this
    will also compute training scores and is merely a utility for plotting the
    results.

    Read more in the :ref:`User Guide <learning_curve>`.

    Parameters
    ----------
    estimator : object type that implements the "fit" and "predict" methods
        An object of that type which is cloned for each validation.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression;
        None for unsupervised learning.

    param_name : string
        Name of the parameter that will be varied.

    param_range : array-like, shape (n_values,)
        The values of the parameter that will be evaluated.

    groups : array-like, with shape (n_samples,), optional
        Group labels for the samples used while splitting the dataset into
        train/test set. Only used in conjunction with a "Group" `cv` instance
        (e.g., `GroupKFold`).

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.20
            ``cv`` default value if None will change from 3-fold to 5-fold
            in v0.22.

    scoring : string, callable or None, optional, default: None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    pre_dispatch : integer or string, optional
        Number of predispatched jobs for parallel execution (default is
        all). The option can reduce the allocated memory. The string can
        be an expression like '2*n_jobs'.

    verbose : integer, optional
        Controls the verbosity: the higher, the more messages.

    error_score : 'raise' | 'raise-deprecating' or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If set to 'raise-deprecating', a FutureWarning is printed before the
        error is raised.
        If a numeric value is given, FitFailedWarning is raised. This parameter
        does not affect the refit step, which will always raise the error.
        Default is 'raise-deprecating' but from version 0.22 it will change
        to np.nan.

    Returns
    -------
    train_scores : array, shape (n_ticks, n_cv_folds)
        Scores on training sets.

    test_scores : array, shape (n_ticks, n_cv_folds)
        Scores on test set.

    Notes
    -----
    See :ref:`sphx_glr_auto_examples_model_selection_plot_validation_curve.py`

    """
    X, y, groups = indexable(X, y, groups)

    cv = check_cv(cv, y, classifier=is_classifier(estimator))
    scorer = check_uplift_scoring(estimator, scoring=scoring)

    parallel = Parallel(n_jobs=n_jobs, pre_dispatch=pre_dispatch,
                        verbose=verbose)
    out = parallel(delayed(_fit_and_score)(
        clone(estimator), X, y, scorer, train, test, verbose,
        parameters={param_name: v}, fit_params=None, return_train_score=True,
        error_score=error_score)
        # NOTE do not change order of iteration to allow one time cv splitters
        for train, test in cv.split(X, y, groups) for v in param_range)
    out = np.asarray(out)
    n_params = len(param_range)
    n_cv_folds = out.shape[0] // n_params
    out = out.reshape(n_cv_folds, n_params, 2).transpose((2, 1, 0))

    return out[0], out[1]


def _aggregate_score_dicts(scores):
    """Aggregate the list of dict to dict of np ndarray

    The aggregated output of _fit_and_score will be a list of dict
    of form [{'prec': 0.1, 'acc':1.0}, {'prec': 0.1, 'acc':1.0}, ...]
    Convert it to a dict of array {'prec': np.array([0.1 ...]), ...}

    Parameters
    ----------

    scores : list of dict
        List of dicts of the scores for all scorers. This is a flat list,
        assumed originally to be of row major order.

    Example
    -------

    >>> scores = [{'a': 1, 'b':10}, {'a': 2, 'b':2}, {'a': 3, 'b':3},
    ...           {'a': 10, 'b': 10}]                         # doctest: +SKIP
    >>> _aggregate_score_dicts(scores)                        # doctest: +SKIP
    {'a': array([1, 2, 3, 10]),
     'b': array([10, 2, 3, 10])}
    """
    return {key: np.asarray([score[key] for score in scores])
            for key in scores[0]}



from ._validation import uplift_check_cv
from ._validation import cross_validate, cross_val_score
from ._validation import cross_val_predict, permutation_test_score
from ._validation import learning_curve
from ._search import GridSearchCV, RandomizedSearchCV
from ..metrics import e_sate, e_satt
from ..metrics import make_uplift_scorer



"""Lenta challenge dataset.

The version used is from the scikit-uplift package with non-ascii
characters removed from the Gender variable.

"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="lenta_dataset.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Lenta/lenta_dataset.csv.gz'),
    checksum=('9002bc5c52ab64c2d68f517cb3390756'
              '87a380d7b407eef1926bd9eda5f31e5d'))

def _float_w_nan(x):
    """Convert strings to floats with empty strings converted to
    nan's."""
    y = [v if v != "" else "nan" for v in x]
    return np.array(y, float), float

def fetch_Lenta(data_home=None, download_if_missing=True,
                random_state=None, shuffle=False,
                categ_as_strings=False, return_X_y=False,
                as_frame=False):
    """Load the Lenta dataset (uplift classification).

    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_response_att : numpy array
        Each value is 1 if store visit occurred 0 otherwise.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_response_att) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['control', 'test']
    gender_values = ["F", "M", "Unspecified", "NA"]
    
    # attribute descriptions
    treatment_descr = [("group", treatment_values)]
    target_descr = [("target_response_att", np.int32, "response_att")]

    feature_descr = [('age', _float_w_nan),
                     ('cheque_count_12m_g20', _float_w_nan),
                     ('cheque_count_12m_g21', _float_w_nan),
                     ('cheque_count_12m_g25', _float_w_nan),
                     ('cheque_count_12m_g32', _float_w_nan),
                     ('cheque_count_12m_g33', _float_w_nan),
                     ('cheque_count_12m_g38', _float_w_nan),
                     ('cheque_count_12m_g39', _float_w_nan),
                     ('cheque_count_12m_g41', _float_w_nan),
                     ('cheque_count_12m_g42', _float_w_nan),
                     ('cheque_count_12m_g45', _float_w_nan),
                     ('cheque_count_12m_g46', _float_w_nan),
                     ('cheque_count_12m_g48', _float_w_nan),
                     ('cheque_count_12m_g52', _float_w_nan),
                     ('cheque_count_12m_g56', _float_w_nan),
                     ('cheque_count_12m_g57', _float_w_nan),
                     ('cheque_count_12m_g58', _float_w_nan),
                     ('cheque_count_12m_g79', _float_w_nan),
                     ('cheque_count_3m_g20',  _float_w_nan),
                     ('cheque_count_3m_g21',  _float_w_nan),
                     ('cheque_count_3m_g25',  _float_w_nan),
                     ('cheque_count_3m_g42',  _float_w_nan),
                     ('cheque_count_3m_g45',  _float_w_nan),
                     ('cheque_count_3m_g52',  _float_w_nan),
                     ('cheque_count_3m_g56',  _float_w_nan),
                     ('cheque_count_3m_g57',  _float_w_nan),
                     ('cheque_count_3m_g79',  _float_w_nan),
                     ('cheque_count_6m_g20',  _float_w_nan),
                     ('cheque_count_6m_g21',  _float_w_nan),
                     ('cheque_count_6m_g25',  _float_w_nan),
                     ('cheque_count_6m_g32',  _float_w_nan),
                     ('cheque_count_6m_g33',  _float_w_nan),
                     ('cheque_count_6m_g38',  _float_w_nan),
                     ('cheque_count_6m_g39',  _float_w_nan),
                     ('cheque_count_6m_g40',  _float_w_nan),
                     ('cheque_count_6m_g41',  _float_w_nan),
                     ('cheque_count_6m_g42',  _float_w_nan),
                     ('cheque_count_6m_g45',  _float_w_nan),
                     ('cheque_count_6m_g46',  _float_w_nan),
                     ('cheque_count_6m_g48',  _float_w_nan),
                     ('cheque_count_6m_g52',  _float_w_nan),
                     ('cheque_count_6m_g56',  _float_w_nan),
                     ('cheque_count_6m_g57',  _float_w_nan),
                     ('cheque_count_6m_g58',  _float_w_nan),
                     ('cheque_count_6m_g79',  _float_w_nan),
                     ('children', _float_w_nan),
                     ('crazy_purchases_cheque_count_12m', _float_w_nan),
                     ('crazy_purchases_cheque_count_1m',  _float_w_nan),
                     ('crazy_purchases_cheque_count_3m',  _float_w_nan),
                     ('crazy_purchases_cheque_count_6m',  _float_w_nan),
                     ('crazy_purchases_goods_count_12m',  _float_w_nan),
                     ('crazy_purchases_goods_count_6m',   _float_w_nan),
                     ('disc_sum_6m_g34', _float_w_nan),
                     ('food_share_15d', _float_w_nan),
                     ('food_share_1m', _float_w_nan),
                     ('gender', gender_values),
                     ('k_var_cheque_15d', _float_w_nan),
                     ('k_var_cheque_3m', _float_w_nan),
                     ('k_var_cheque_category_width_15d', _float_w_nan),
                     ('k_var_cheque_group_width_15d', _float_w_nan),
                     ('k_var_count_per_cheque_15d_g24', _float_w_nan),
                     ('k_var_count_per_cheque_15d_g34', _float_w_nan),
                     ('k_var_count_per_cheque_1m_g24', _float_w_nan),
                     ('k_var_count_per_cheque_1m_g27', _float_w_nan),
                     ('k_var_count_per_cheque_1m_g34', _float_w_nan),
                     ('k_var_count_per_cheque_1m_g44', _float_w_nan),
                     ('k_var_count_per_cheque_1m_g49', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g24', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g27', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g32', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g34', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g41', _float_w_nan),
                     ('k_var_count_per_cheque_3m_g44', _float_w_nan),
                     ('k_var_count_per_cheque_6m_g24', _float_w_nan),
                     ('k_var_count_per_cheque_6m_g27', _float_w_nan),
                     ('k_var_count_per_cheque_6m_g32', _float_w_nan),
                     ('k_var_count_per_cheque_6m_g44', _float_w_nan),
                     ('k_var_days_between_visits_15d', _float_w_nan),
                     ('k_var_days_between_visits_1m', _float_w_nan),
                     ('k_var_days_between_visits_3m', _float_w_nan),
                     ('k_var_disc_per_cheque_15d', _float_w_nan),
                     ('k_var_disc_share_12m_g32', _float_w_nan),
                     ('k_var_disc_share_15d_g24', _float_w_nan),
                     ('k_var_disc_share_15d_g34', _float_w_nan),
                     ('k_var_disc_share_15d_g49', _float_w_nan),
                     ('k_var_disc_share_1m_g24', _float_w_nan),
                     ('k_var_disc_share_1m_g27', _float_w_nan),
                     ('k_var_disc_share_1m_g34', _float_w_nan),
                     ('k_var_disc_share_1m_g40', _float_w_nan),
                     ('k_var_disc_share_1m_g44', _float_w_nan),
                     ('k_var_disc_share_1m_g49', _float_w_nan),
                     ('k_var_disc_share_1m_g54', _float_w_nan),
                     ('k_var_disc_share_3m_g24', _float_w_nan),
                     ('k_var_disc_share_3m_g26', _float_w_nan),
                     ('k_var_disc_share_3m_g27', _float_w_nan),
                     ('k_var_disc_share_3m_g32', _float_w_nan),
                     ('k_var_disc_share_3m_g33', _float_w_nan),
                     ('k_var_disc_share_3m_g34', _float_w_nan),
                     ('k_var_disc_share_3m_g38', _float_w_nan),
                     ('k_var_disc_share_3m_g40', _float_w_nan),
                     ('k_var_disc_share_3m_g41', _float_w_nan),
                     ('k_var_disc_share_3m_g44', _float_w_nan),
                     ('k_var_disc_share_3m_g46', _float_w_nan),
                     ('k_var_disc_share_3m_g48', _float_w_nan),
                     ('k_var_disc_share_3m_g49', _float_w_nan),
                     ('k_var_disc_share_3m_g54', _float_w_nan),
                     ('k_var_disc_share_6m_g24', _float_w_nan),
                     ('k_var_disc_share_6m_g27', _float_w_nan),
                     ('k_var_disc_share_6m_g32', _float_w_nan),
                     ('k_var_disc_share_6m_g34', _float_w_nan),
                     ('k_var_disc_share_6m_g44', _float_w_nan),
                     ('k_var_disc_share_6m_g46', _float_w_nan),
                     ('k_var_disc_share_6m_g49', _float_w_nan),
                     ('k_var_disc_share_6m_g54', _float_w_nan),
                     ('k_var_discount_depth_15d', _float_w_nan),
                     ('k_var_discount_depth_1m', _float_w_nan),
                     ('k_var_sku_per_cheque_15d', _float_w_nan),
                     ('k_var_sku_price_12m_g32', _float_w_nan),
                     ('k_var_sku_price_15d_g34', _float_w_nan),
                     ('k_var_sku_price_15d_g49', _float_w_nan),
                     ('k_var_sku_price_1m_g24', _float_w_nan),
                     ('k_var_sku_price_1m_g26', _float_w_nan),
                     ('k_var_sku_price_1m_g27', _float_w_nan),
                     ('k_var_sku_price_1m_g34', _float_w_nan),
                     ('k_var_sku_price_1m_g40', _float_w_nan),
                     ('k_var_sku_price_1m_g44', _float_w_nan),
                     ('k_var_sku_price_1m_g49', _float_w_nan),
                     ('k_var_sku_price_1m_g54', _float_w_nan),
                     ('k_var_sku_price_3m_g24', _float_w_nan),
                     ('k_var_sku_price_3m_g26', _float_w_nan),
                     ('k_var_sku_price_3m_g27', _float_w_nan),
                     ('k_var_sku_price_3m_g32', _float_w_nan),
                     ('k_var_sku_price_3m_g33', _float_w_nan),
                     ('k_var_sku_price_3m_g34', _float_w_nan),
                     ('k_var_sku_price_3m_g40', _float_w_nan),
                     ('k_var_sku_price_3m_g41', _float_w_nan),
                     ('k_var_sku_price_3m_g44', _float_w_nan),
                     ('k_var_sku_price_3m_g46', _float_w_nan),
                     ('k_var_sku_price_3m_g48', _float_w_nan),
                     ('k_var_sku_price_3m_g49', _float_w_nan),
                     ('k_var_sku_price_3m_g54', _float_w_nan),
                     ('k_var_sku_price_6m_g24', _float_w_nan),
                     ('k_var_sku_price_6m_g26', _float_w_nan),
                     ('k_var_sku_price_6m_g27', _float_w_nan),
                     ('k_var_sku_price_6m_g32', _float_w_nan),
                     ('k_var_sku_price_6m_g41', _float_w_nan),
                     ('k_var_sku_price_6m_g42', _float_w_nan),
                     ('k_var_sku_price_6m_g44', _float_w_nan),
                     ('k_var_sku_price_6m_g48', _float_w_nan),
                     ('k_var_sku_price_6m_g49', _float_w_nan),
                     ('main_format', _float_w_nan),
                     ('mean_discount_depth_15d', _float_w_nan),
                     ('months_from_register', _float_w_nan),
                     ('perdelta_days_between_visits_15_30d', _float_w_nan),
                     ('promo_share_15d', _float_w_nan),
                     ('response_sms', _float_w_nan),
                     ('response_viber', _float_w_nan),
                     ('sale_count_12m_g32', _float_w_nan),
                     ('sale_count_12m_g33', _float_w_nan),
                     ('sale_count_12m_g49', _float_w_nan),
                     ('sale_count_12m_g54', _float_w_nan),
                     ('sale_count_12m_g57', _float_w_nan),
                     ('sale_count_3m_g24', _float_w_nan),
                     ('sale_count_3m_g33', _float_w_nan),
                     ('sale_count_3m_g57', _float_w_nan),
                     ('sale_count_6m_g24', _float_w_nan),
                     ('sale_count_6m_g25', _float_w_nan),
                     ('sale_count_6m_g32', _float_w_nan),
                     ('sale_count_6m_g33', _float_w_nan),
                     ('sale_count_6m_g44', _float_w_nan),
                     ('sale_count_6m_g54', _float_w_nan),
                     ('sale_count_6m_g57', _float_w_nan),
                     ('sale_sum_12m_g24', _float_w_nan),
                     ('sale_sum_12m_g25', _float_w_nan),
                     ('sale_sum_12m_g26', _float_w_nan),
                     ('sale_sum_12m_g27', _float_w_nan),
                     ('sale_sum_12m_g32', _float_w_nan),
                     ('sale_sum_12m_g44', _float_w_nan),
                     ('sale_sum_12m_g54', _float_w_nan),
                     ('sale_sum_3m_g24', _float_w_nan),
                     ('sale_sum_3m_g26', _float_w_nan),
                     ('sale_sum_3m_g32', _float_w_nan),
                     ('sale_sum_3m_g33', _float_w_nan),
                     ('sale_sum_6m_g24', _float_w_nan),
                     ('sale_sum_6m_g25', _float_w_nan),
                     ('sale_sum_6m_g26', _float_w_nan),
                     ('sale_sum_6m_g32', _float_w_nan),
                     ('sale_sum_6m_g33', _float_w_nan),
                     ('sale_sum_6m_g44', _float_w_nan),
                     ('sale_sum_6m_g54', _float_w_nan),
                     ('stdev_days_between_visits_15d', _float_w_nan),
                     ('stdev_discount_depth_15d', _float_w_nan),
                     ('stdev_discount_depth_1m', _float_w_nan),
                     ]


    ret = _fetch_remote_csv(ARCHIVE, "Lenta",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=195
                            )
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""Utility functions for statistical computations."""

import numpy as np

def quantile(a, q, axis=None, out=None, overwrite_input=False,
             method='linear', keepdims=False, *, weights=None,
             interpolation=None):
    """Weighted quantiles functions.

    When sample_weight is None np.quantile is used.  Also used for
    numpy >= 2.0 since inverse_cdf strategy gives strange results for
    small samples.

    Warning: results for sample_weight=None and
    sample_weight=[1,...,1] might not be identical since a different
    strategy.

    """
    if weights is None:
        return np.quantile(a, q, axis=axis, out=out,
                           overwrite_input=overwrite_input,
                           method=method, keepdims=keepdims,
                           interpolation=interpolation)

    def weighted_quantile(data, quantiles, weights):
        """
        from https://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy
        """
        ix = np.argsort(data)
        data = data[ix] # sort data
        weights = weights[ix] # sort weights
        cdf = (np.cumsum(weights) - 0.5 * weights) / np.sum(weights) # 'like' a CDF function
        q = np.interp(quantiles, cdf, data)
        print(data, quantiles, weights, q)
        return q
    arr = np.asanyarray(a)
    return weighted_quantile(arr, q, weights)





"""Handle class imbalance."""

import numpy as np

from sklearn.utils import check_random_state

from ..base import UpliftRegressorMixin

def class_flip(y, trt, n_trt=None, kc0=1, kc1=1, kt0=1, kt1=1, k=None,
               random_state=None):
    rng = check_random_state(random_state)
    if k is not None:
        raise RuntimeError("multiclass/treatment flipping not implemented.")
    assert 0 <= kc0 <= 1
    assert 0 <= kc1 <= 1
    assert 0 <= kt0 <= 1
    assert 0 <= kt1 <= 1
    mask_c0 = ((trt==0) & (y==0))
    mask_c1 = ((trt==0) & (y==1))
    mask_t0 = ((trt==1) & (y==0))
    mask_t1 = ((trt==1) & (y==1))
    for mask, k in [(mask_c0, kc0), (mask_c1, kc1), (mask_t0, kt0), (mask_t1, kt1)]:
        if k == 1:
            continue
        n = mask.sum()
        n_flip = int(n * k)
        mask_flip = rng.choice(np.flatnonzero(mask), n_flip)
        y[mask_flip] = 1-y[mask_flip]
    return y




from pytest import approx

import numpy as np

from sklearn.linear_model import LogisticRegression

from usklearn.classifiers import ClassifierAsRegressor

def _make_data():
    X = [[1.0], [1], [1], [0], [0], [0]]
    y = [1,1,0, 0,0,1]
    return X, y
    
def test_classifier_as_regressor():
    lr = LogisticRegression(penalty=None)
    lr_reg = ClassifierAsRegressor(lr)
    X, y = _make_data()
    lr_reg.fit(X, y)
    
    pred_y = lr_reg.predict(X)
    assert pred_y[:3] == approx(2/3, abs=1e-3)
    assert pred_y[3:] == approx(1/3, abs=1e-3)

def test_classifier_as_regressor_pos_label():
    lr = LogisticRegression(penalty=None)
    lr_reg = ClassifierAsRegressor(lr, pos_label=0)
    X, y = _make_data()
    lr_reg.fit(X, y)
    
    pred_y = lr_reg.predict(X)
    assert pred_y[:3] == approx(1/3, abs=1e-3)
    assert pred_y[3:] == approx(2/3, abs=1e-3)

def test_classifier_as_regressor_log_proba():
    lr = LogisticRegression(penalty=None)
    lr_reg = ClassifierAsRegressor(lr, response_method="predict_log_proba")
    X, y = _make_data()
    lr_reg.fit(X, y)
    
    pred_y = lr_reg.predict(X)
    assert pred_y[:3] == approx(np.log(2/3), abs=1e-3)
    assert pred_y[3:] == approx(np.log(1/3), abs=1e-3)

def test_classifier_as_regressor_decision_function():
    lr = LogisticRegression(penalty=None)
    lr_reg = ClassifierAsRegressor(lr, response_method="decision_function")
    X, y = _make_data()
    lr_reg.fit(X, y)
    
    pred_y = lr_reg.predict(X)
    assert pred_y[:3] == approx(np.log(2), abs=1e-3)
    assert pred_y[3:] == approx(np.log(1/2), abs=1e-3)



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 13:19:24 2022

@author: Krzysztof Rudaś
"""

"""Uplift models based on multiple classification/regression models."""

import numpy as np
from scipy.linalg.lapack import get_lapack_funcs, _compute_lwork
from scipy.linalg.misc import LinAlgError, LinAlgWarning
from scipy import linalg
from scipy.linalg import lstsq
import scipy.sparse as sp
from sklearn.base import BaseEstimator, clone
from sklearn.utils import check_X_y, check_consistent_length
from sklearn.neural_network import MLPRegressor
from sklearn.utils.metaestimators import _BaseComposition
from ..base import UpliftRegressorMixin
from ..utils import check_trt

class MLPUpliftRegressor(_BaseComposition, UpliftRegressorMixin):
    """Multimodel uplift regressor.

    Build separate models for control and all treatments, subtract
    control predictions from treatment predictions.

    Parameters
    ----------

    base_estimator : a sklearn regressor or list of (name, regressor)
        tuples.  If a list is provided the first model is used for
        control, successive one for treatments.  If different parameters are
        to be used per each model, the list version must be given.  If a
        single estimator is given it will be cloned for every treatment.
    """
    def __init__(self, base_estimator=MLPRegressor(hidden_layer_sizes=(20,20),max_iter = 2000)):
        self.base_estimator = base_estimator
    def _check_base_estimator(self, n_models):
        if hasattr(self.base_estimator, "fit"):
            estimator_list = []
            for i in range(n_models):
                name = "model_c" if i == 0 else "model_t" + str(i-1)
                estimator_list.append((name, clone(self.base_estimator)))
        else:
            estimator_list = self.base_estimator
        return estimator_list
    def fit(self, X, y, trt, n_trt=None):
        X, y = check_X_y(X, y, accept_sparse="csr")
        self.trt_, self.n_trt_ = check_trt(trt, n_trt)
        check_consistent_length(X, y, self.trt_)
        self.n_models_ = self.n_trt_ + 1
        self.models_ = self._check_base_estimator(self.n_models_)
        #import pdb; pdb.set_trace()
        for i in range(self.n_models_):
            mi = self.models_[i][1]
            ind = (trt==i)
            Xi = X[ind]
            yi = y[ind]
            mi.fit(Xi, yi)
        return self
    def predict(self, X):
        y_control = self.models_[0][1].predict(X)
        cols = [self.models_[i+1][1].predict(X) - y_control
                    for i in range(self.n_trt_)]
        if self.n_trt_ == 1:
            y = cols[0]
        else:
            y = np.column_stack(cols)
        return y
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        if hasattr(self.base_estimator, "fit"):
            return super().get_params(deep=deep)
        return self._get_params('base_estimator', deep=deep)

    def set_params(self, **kwargs):
        """Set the parameters of this estimator.
        Valid parameter keys can be listed with ``get_params()``.
        Returns
        -------
        self
        """
        if hasattr(self.base_estimator, "fit"):
            return super().set_params(**kwargs)
        else:
            self._set_params('base_estimator', **kwargs)
        return self



"""Meta models: uplift models combined of classifiers/regressors."""

from .multi_model import MultimodelUpliftRegressor
from .multi_model import MultimodelUpliftLinearRegressor
from .multi_model import MultimodelUpliftClassifier
from .response import TreatmentUpliftClassifier
from .response import ResponseUpliftClassifier
from .response import ControlUpliftClassifier
from .target_transform import TargetTransformUpliftRegressor
from .target_transform import TargetTransformUpliftClassifier
from .s_learner import SLearnerUpliftRegressor
from .s_learner import SLearnerUpliftClassifier
from .nested import NestedMeanUpliftRegressor
from .nested import DDRUpliftClassifier
from .x_learner import XLearnerUpliftRegressor

__all__ = ["MultimodelUpliftRegressor",
           "MultimodelUpliftLinearRegressor",
           "MultimodelUpliftClassifier",
           "TreatmentUpliftClassifier",
           "ResponseUpliftClassifier",
           "ControlUpliftClassifier",
           "TargetTransformUpliftRegressor",
           "TargetTransformUpliftClassifier",
           "SLearnerUpliftRegressor",
           "SLearnerUpliftClassifier",
           "NestedMeanUpliftRegressor",
           "DDRUpliftClassifier",
           ]



import numpy as np
import scipy.sparse as sp

def safe_hstack(Xs):
    """hstack which works for dense and sparse arrays."""
    if any(sp.issparse(f) for f in Xs):
        Xs = sp.hstack(Xs, format="csr")
    else:
        Xs = np.hstack(Xs)
    return Xs



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 22 15:12:29 2021

@author: Krzysztof Rudaś
"""

"""Uplift models based on multiple classification/regression models."""

import numpy as np
from scipy.linalg.lapack import get_lapack_funcs, _compute_lwork
from scipy.linalg.misc import LinAlgError, LinAlgWarning
from scipy import linalg
from scipy.linalg import lstsq
import scipy.sparse as sp
from sklearn.base import BaseEstimator, clone
from sklearn.utils import check_X_y, check_consistent_length
from sklearn.linear_model import LinearRegression
from sklearn.linear_model._base import LinearModel
from sklearn.utils.metaestimators import _BaseComposition
from ..base import UpliftRegressorMixin
from ..utils import check_trt

def Gaussian(x,Sigma):
    return -x@np.linalg.inv(Sigma)@x.T

def GaussianMatrix(X,y,Sigma):
    row,col=X.shape
    G_x=np.zeros(shape=(row,col))
    G_y=np.zeros(shape=row)
    g = np.zeros(shape=row)
    X=np.asarray(X)
    GG = Gaussian(X,Sigma)
    j=0
    for v_j in X:
        gg = np.exp(np.diag(GG)+GG[j,j]-2*GG[:,j])
        g = g+gg
        G_x=G_x+gg[None].T@v_j[None]
        G_y=G_y+gg*y[j]
        j+=1
    G_x=G_x/g[:,None]
    G_y=G_y/g
    return (G_x,G_y)


def GaussianMatrix2(X,y,Sigma,beta):
    row,col=X.shape
    G_g=np.zeros(shape=row)
    g = np.zeros(shape=row)
    X=np.asarray(X)
    GG = Gaussian(X,Sigma)
    j=0
    for v_j in X:
        gg = np.exp(np.diag(GG)+GG[j,j]-2*GG[:,j])
        g = g+gg
        G_g=G_g+gg*(y[j]-v_j.T@beta)
        j+=1
    G_g=G_g/g
    return (G_g)


class MultimodelUpliftRegressorRobinson(_BaseComposition, UpliftRegressorMixin):
    """Multimodel uplift regressor.

    Build separate models for control and all treatments, subtract
    control predictions from treatment predictions.

    Parameters
    ----------

    base_estimator : a sklearn regressor or list of (name, regressor)
        tuples.  If a list is provided the first model is used for
        control, successive one for treatments.  If different parameters are
        to be used per each model, the list version must be given.  If a
        single estimator is given it will be cloned for every treatment.
    """
    def __init__(self, base_estimator=LinearRegression()):
        self.base_estimator = base_estimator
    def _check_base_estimator(self, n_models):
        if hasattr(self.base_estimator, "fit"):
            estimator_list = []
            for i in range(n_models):
                name = "model_c" if i == 0 else "model_t" + str(i-1)
                estimator_list.append((name, clone(self.base_estimator)))
        else:
            estimator_list = self.base_estimator
        return estimator_list
    def fit(self, X, y, trt, n_trt=None):
        X, y = check_X_y(X, y, accept_sparse="csr")
        self.trt_, self.n_trt_ = check_trt(trt, n_trt)
        check_consistent_length(X, y, self.trt_)
        self.n_models_ = self.n_trt_ + 1
        self.models_ = self._check_base_estimator(self.n_models_)
        self.n_ = np.empty(self.n_models_, dtype=int)
        self.p = X.shape[1]
        self.sigma = np.empty(self.n_models_)
        #import pdb; pdb.set_trace()
        for i in range(self.n_models_):
            mi = self.models_[i][1]
            ind = (trt==i)
            self.n_[i] = ind.sum()
            Xi = X[ind]
            yi = y[ind]
            Sigma = np.eye(self.p)
            M= GaussianMatrix(Xi,yi,3*Sigma)
            #mi.fit(Xi-M[0], yi-M[1])
            mi.fit(Xi, yi)
            mi.coef_=np.linalg.inv((Xi-M[0]).T@(Xi-M[0]))@(Xi-M[0]).T@(yi-M[1])
            #mi.intercept_=0
            #g = GaussianMatrix2(Xi,yi,3*Sigma,mi.coef_)
            #mi.fit(Xi, yi-g)
            #mi.coef_
        return self
    def predict(self, X):
        y_control = self.models_[0][1].predict(X)
        cols = [self.models_[i+1][1].predict(X) - y_control
                    for i in range(self.n_trt_)]
        if self.n_trt_ == 1:
            y = cols[0]
        else:
            y = np.column_stack(cols)
        return y
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        if hasattr(self.base_estimator, "fit"):
            return super().get_params(deep=deep)
        return self._get_params('base_estimator', deep=deep)

    def set_params(self, **kwargs):
        """Set the parameters of this estimator.
        Valid parameter keys can be listed with ``get_params()``.
        Returns
        -------
        self
        """
        if hasattr(self.base_estimator, "fit"):
            return super().set_params(**kwargs)
        else:
            self._set_params('base_estimator', **kwargs)
        return self


class MultimodelUpliftLinearRegressorRobinson(MultimodelUpliftRegressorRobinson, LinearModel):
    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self._set_coef()
        return self
    def _set_coef(self):
        if not hasattr(self.models_[0][1], "coef_"):
            raise RuntimeError("Base estimator for multi-linear"
                                   " model must set coef_ attribute")
        c0 = self.models_[0][1].coef_
        i0 = self.models_[0][1].intercept_
        coef = np.empty((self.n_trt_, len(c0)))
        intercept = np.empty(self.n_trt_)
        for i in range(self.n_trt_):
            ui = self.models_[i+1][1].coef_ - c0
            #print(c0)
            ii = self.models_[i+1][1].intercept_ - i0
            coef[i,:] = ui
            intercept[i] = ii
        if self.n_trt_ == 1:
            coef = np.squeeze(coef)
            intercept = np.squeeze(intercept)
        self.coef_ = coef
        self.intercept_ = intercept
    def predict(self, X):
        return LinearModel.predict(self, X)



"""Metrics to assess performance of uplift regression task

Functions named as ``*_score`` return a scalar value to maximize: the higher
the better

Function named as ``*_error`` or ``*_loss`` return a scalar value to minimize:
the lower the better
"""

import numpy as np

from sklearn.utils.validation import check_array, check_consistent_length

from ..utils.validation import check_trt

def _e_satx(y_true, y_pred, trt, n_trt=None, satt=True):
    """Generic method for e_sate and e_satt."""
    y_true = check_array(y_true, ensure_2d=False)
    y_pred = check_array(y_pred, ensure_2d=False)
    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))
    trt, n_trt = check_trt(trt, n_trt)
    check_consistent_length(y_true, y_pred, trt)

    average_ys_true = []
    sate_pred_s = []
    nts = []
    for t in range(n_trt + 1):
        ind = (trt==t)
        nt = ind.sum()
        if nt > 0:
            av_y_t_true = np.average(y_true[ind])
            if t > 0:
                if satt:
                    sate_pred = np.average(y_pred[ind][:,t-1]) # ATT
                else:
                    sate_pred = np.average(y_pred[:,t-1]) # ATE
            else:
                sate_pred = np.nan
        else:
            av_y_t_true = np.nan
            sate_pred = np.nan
        nts.append(nt)
        average_ys_true.append(av_y_t_true)
        sate_pred_s.append(sate_pred)
    if nts[0] == 0:
        raise RuntimeError("Cannot estimate ATE. No cases in control")
    n_treated = sum(nts[1:])
    if n_treated == 0:
        raise RuntimeError("Cannot estimate ATE. No cases in any treatment")
    average_y_control = average_ys_true[0]
    sate_pred_s = sate_pred_s[1:]
    sate_true_s = [average_ys_true[t+1] - average_y_control
                       for t in range(n_trt)]
    return sum(nt/n_treated * abs(sate_pred - sate_true)
                   for nt, sate_pred, sate_true in
                   zip(nts[1:], sate_pred_s, sate_true_s))

def e_sate(y_true, y_pred, trt, n_trt=1):
    """Absolute error on Sample Average Treatment Effect.

    Works by computing ATE using model predictions and true outcomes.
    Absolute value of the difference between the two values is returned.

    For multiple treatments, return weighted average.  This measure is
    not very effective.
    """
    return _e_satx(y_true, y_pred, trt, n_trt=n_trt, satt=False)
def e_satt(y_true, y_pred, trt, n_trt=1):
    """Absolute error on Sample Average Treatment Effect on the Treated.

    Works by computing ATT using model predictions and true outcomes.
    Absolute value of the difference between the two values is returned.

    For multiple treatments, return weighted average.  This measure is
    not very effective.
    """
    return _e_satx(y_true, y_pred, trt, n_trt=n_trt, satt=True)

def sim_gain_at(y_true, y_pred, trt, n_trt=None, *, quantile=0.5):
    """Simulated on-policy gains at given quantile of targeted
    population.

    For mulitple treatments, the most beneficial treatment of each
    case is selected.  Enough data must be present in all treatments
    for the measure to be well defined.
    """
    y_true = check_array(y_true, ensure_2d=False)
    y_pred = check_array(y_pred, ensure_2d=False)
    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))
    trt, n_trt = check_trt(trt, n_trt)
    check_consistent_length(y_true, y_pred, trt)

    best_treatment = y_pred.argmax(axis=1)
    best_gain = y_pred[:,best_treatment]
    
    
    average_ys_true = []
    sate_pred_s = []
    nts = []
    for t in range(n_trt + 1):
        ind = (trt==t)
        nt = ind.sum()
        if nt > 0:
            av_y_t_true = np.average(y_true[ind])
            if t > 0:
                if satt:
                    sate_pred = np.average(y_pred[ind][:,t-1]) # ATT
                else:
                    sate_pred = np.average(y_pred[:,t-1]) # ATE
            else:
                sate_pred = np.nan
        else:
            av_y_t_true = np.nan
            sate_pred = np.nan
        nts.append(nt)
        average_ys_true.append(av_y_t_true)
        sate_pred_s.append(sate_pred)
    if nts[0] == 0:
        raise RuntimeError("Cannot estimate ATE. No cases in control")
    n_treated = sum(nts[1:])
    if n_treated == 0:
        raise RuntimeError("Cannot estimate ATE. No cases in any treatment")
    average_y_control = average_ys_true[0]
    sate_pred_s = sate_pred_s[1:]
    sate_true_s = [average_ys_true[t+1] - average_y_control
                       for t in range(n_trt)]
    return sum(nt/n_treated * abs(sate_pred - sate_true)
                   for nt, sate_pred, sate_true in
                   zip(nts[1:], sate_pred_s, sate_true_s))

def sim_gain(y_true, y_pred, trt, n_trt=None, *, quantile=0.5):
    """Simulated on-policy gains for the whole population.
    """
    return sim_gain_at(y_true, y_pred, trt, n_trt=None, quantile=1.1)



"""An A/B testing dataset from Kaggle.
"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE = RemoteFileMetadata(
    filename="marketing_AB.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/releases/download/marketing_AB/marketing_AB.csv.gz'),
    checksum=('a318767a73785b7e54fbf36cbf300411'
              '5f61e943844b7ac8d3753a893e093d9d'))

def fetch_marketing_AB(data_home=None, download_if_missing=True,
                    random_state=None, shuffle=False,
                    categ_as_strings=False, return_X_y=False,
                    as_frame=False):
    """Load the marketing_AB dataset from Kaggle.
    Download it if necessary.
    
    See https://www.kaggle.com/datasets/faviovaz/marketing-ab-testing
    for details.

    The treatment was showing the user an advertisement ('ad'), the
    control showing a Public Service Announcement ('psa').

    The dataset exhibits very high class and treatment imbalance.

    Changes made to the original dataset:
     * removed record number column
     * changed spaces to _ in column names
     * changed target from bool to {0,1}


    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_purchase : numpy array
        Indicator whether a purchase was made.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_purchase) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['psa', 'ad']
    most_ads_day_values = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']

    # attribute descriptions
    treatment_descr = [("test_group", treatment_values)]
    target_descr = [("target_converted", np.int32, "converted"),]

    feature_descr = [("user_id", float),
                     ("total_ads", float),
                     ("most_ads_day", most_ads_day_values),
                     ("most_ads_hour", np.int32),
    ]

    ret = _fetch_remote_csv(ARCHIVE, "marketing_AB",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=6,
                            )
    if not return_X_y:
        ret.descr = __doc__

    return ret



import numpy as np

from usklearn.metrics.bins import iter_quantiles

def test_iter_quantiles():
    trt =    np.array([0,0,0,0,0, 1,1,1,1,1])
    scores = np.array([1,2,3,4,5, 5,4,3,2,1], dtype=float)

    for i, q in enumerate(iter_quantiles(scores, trt, 1, n=5)):
        assert len(q) == 2
        assert q[0].shape == (1,)
        assert q[1].shape == (1,)
        assert q[0].item() == (4-i)
        assert q[1].item() == 9-(4-i)

def test_iter_quantiles_weighted():
    trt =     np.array([0,0,0,0,0, 1,1,1,1,1])
    scores =  np.array([1,2,3,4,5, 5,4,3,2,1], dtype=float)
    weights = np.array([1,1,1,1,1, 1,1,1,1,1], dtype=float)

    for i, q in enumerate(iter_quantiles(scores, trt, 1, n=5,
                                         sample_weight=weights)):
        assert len(q) == 2
        assert q[0].shape == (1,)
        assert q[1].shape == (1,)
        assert q[0].item() == (4-i)
        assert q[1].item() == 9-(4-i)

def test_iter_quantiles_weighted2():
    trt =     np.array([0,0,0,0,0, 1,1,1,1,1])
    scores =  np.array([1,2,3,4,5, 5,4,3,2,1], dtype=float)
    weights = np.array([1,1,0,0,0, 0,0,0,1,1], dtype=float)

    for i, q in enumerate(iter_quantiles(scores, trt, 1, n=2,
                                         sample_weight=weights)):
        assert len(q) == 2
        if i == 0:
            assert q[0].shape == (4,)
            assert q[1].shape == (4,)
            assert set(q[0]) == set([1,2,3,4])
            assert set(q[1]) == set([5,6,7,8])
        elif i == 1:
            assert q[0].shape == (1,)
            assert q[1].shape == (1,)
            assert set(q[0]) == set([0])
            assert set(q[1]) == set([9])
        else:
            assert False



import numpy as np

from sklearn.model_selection import train_test_split

from usklearn.utils.multi_array import MultiArray

def _prepare_data():
    n = 100
    p = 10
    X = np.repeat(np.arange(n, dtype=float).reshape(-1,1), p, axis=1)
    y = np.arange(n)
    w = np.arange(n, dtype=float)
    trt = np.arange(n) % 3
    n_trt = 3
    m = MultiArray(X, {"w":w, "trt":trt}, {"n_trt":n_trt})
    return m    
def test_multiarray1():
    """Test simple indexing."""
    m = _prepare_data()
    # test ranges and boolean indices
    for s in [slice(1,2), slice(50,100), (np.arange(m.shape[0]) % 2 == 1)]:
        ms = m[s]
        assert ms.scalar_dict == m.scalar_dict
        if isinstance(s, slice):
            r = np.arange(s.start, s.stop, s.step, dtype=float)
        else:
            r = np.arange(m.shape[0], dtype=float)[s]
        assert np.all(ms.main_array == r.reshape(-1,1))
        assert np.all(ms.array_dict["trt"] == r.astype(int) % 3)
        assert np.all(ms.array_dict["w"] == r)
        assert ms.shape[0] == r.shape[0]
        assert ms.shape[1] == m.shape[1]
def test_multiarray2():
    """Test crossvalidation."""
    m = _prepare_data()
    # test crossvalidation
    m1, m2 = train_test_split(m)
    w = set(m1.array_dict["w"]) | set(m2.array_dict["w"])
    w = list(w)
    w.sort()
    assert np.all(np.array(w) == np.arange(m.shape[0]))



"""Uplift models based on target transform."""

import numpy as np

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression

from .base import UpliftMetaModelBase
from ..base import UpliftRegressorMixin
from ..base import UpliftClassifierMixin


class _TargetTransformUpliftModelBase(UpliftMetaModelBase):
    def _get_model_names_list(self, X=None, y=None, trt=None):
        m_names = []
        for i in range(self.n_trt_):
            name = "model_ct"
            if self.n_trt_ > 1:
                name += str(i)
            m_names.append(name)
        return m_names
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        c_mask = (trt==0)
        for i in range(self.n_models_):
            t_mask = (trt==(i+1))
            mask = t_mask|c_mask
            if sample_weight is None:
                w_i = None
            else:
                w_i = sample_weight[mask]
            X_i = X[mask]
            y_i = y[mask]
            trt_i = trt[mask]
            X_i, y_i, w_i = self._transform(X_i, y_i, trt_i, n_trt, w_i, y)
            yield X_i, y_i, w_i

    def _transform(self, X, y, trt, n_trt, sample_weight, full_y):
        """Transform target for model building.

        full_y is passed to allow tests to avoid overwriting."""
        raise NotImplementedError()

class TargetTransformUpliftRegressor(UpliftRegressorMixin,
                                     _TargetTransformUpliftModelBase):
    def __init__(self, base_estimator=LinearRegression()):
        super().__init__(base_estimator=base_estimator)
    def _transform(self, X, y, trt, n_trt, sample_weight, full_y):
        """Transform target for model building.

        full_y is passed to allow tests to avoid overwriting."""
        n = trt.shape[0]
        mask_c = (trt==0)
        mask_t = ~mask_c
        nt = mask_t.sum()
        nc = mask_c.sum()
        n = nt + nc
        y = np.asarray(y, float) # allow classification problems
        if np.may_share_memory(y, full_y):
            y = y.copy()
        y[mask_c] *= (-n/nc)
        y[mask_t] *= (n/nt)
        return X, y, sample_weight
    def predict(self, X):
        preds = [m_i.predict(X) for _, m_i in self.models_]
        if self.n_trt_ == 1:
            y = preds[0]
        else:
            y = np.column_stack(preds)
        return y


class TargetTransformUpliftClassifier(UpliftClassifierMixin,
                                      _TargetTransformUpliftModelBase):
    def __init__(self, base_estimator=LogisticRegression()):
        super().__init__(base_estimator=base_estimator)
    def _transform(self, X, y, trt, n_trt, sample_weight, full_y):
        """Transform target for model building.

        full_y is passed to allow tests to avoid overwriting."""
        if np.may_share_memory(y, full_y):
            y = y.copy()
        y[trt == 0] = 1-y[trt == 0]
        return X, y, sample_weight
    def predict(self, X):
        preds = [2 * m_i.predict_proba(X) - 1 for _, m_i in self.models_]
        if self.n_trt_ == 1:
            y = preds[0]
        else:
            y = np.dstack(preds)
        return y



"""Lazada E-comerce dataset.

See https://github.com/kailiang-zhong/DESCN/tree/main/data/Lazada_dataset/ for details.
"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE_TRAIN = RemoteFileMetadata(
    filename="Lazada_train.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Lazada/Lazada_trainset.csv.gz'),
    checksum=('5a46ec368f9e8397267e818447ea7f3d'
              '239a671731ca98c0ae9f17c19ab5a469'))
ARCHIVE_TEST = RemoteFileMetadata(
    filename="Lazada_test.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Lazada/Lazada_testset.csv.gz'),
    checksum=('a6d8832a71e8f0b6d4e8858d91ce7de9'
              'd8e0f2f58aa7208c78b1e0f7d6ba9c94'))


def fetch_Lazada(version="test", data_home=None, download_if_missing=True,
                 random_state=None, shuffle=False,
                 categ_as_strings=False, return_X_y=False,
                 as_frame=False):
    """Load the Lazada e-comerce dataset.

    Download it if necessary.

    There is a training and test set available.  The test dataset
    comes from a randomized experiment, in the training set assignment
    is biased.
    
    Details of the dataset can be found in the following paper:
    https://arxiv.org/pdf/2207.09920.

    The license is available at https://github.com/jszymon/uplift_sklearn_data/releases/download/Lazada/Lazada_LICENSE

    Parameters
    ----------
    version : string, optional
        Specify whether to return training (``train'') or testing
        (``test'') dataset.  Test dataset comes from a randomized
        experiment, in the training set assignment is biased.

    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target_visit : numpy array
        Each value is 1 if website visit occurred 0 otherwise.

    dataset.target_conversion : numpy array
        Each value is 1 if purchase occurred 0 otherwise.

    dataset.target_spend : numpy array
        Each value corresponds to the amount of money spent.

    dataset.DESCR : string
        Description of the Hillstrom dataset.

    (data, target_conversion, target_visit, target_exposure) : tuple if
        ``return_X_y`` is True

    """

    if version == "train":
        ARCHIVE = ARCHIVE_TRAIN
        file_name = "Lazada_train"
    elif version == "test":
        ARCHIVE = ARCHIVE_TEST
        file_name = "Lazada_test"
    else:
        raise ValueError(f"Wrong version ({version}) of Lazada dataset"
                         "requested.\nValid choices are `train' or `test'")

    # dictionaries
    treatment_values = ['control', 'treated']
    categ_values = dict()

    # attribute descriptions
    treatment_descr = [("is_treat", np.int32)]
    target_descr = [("target", np.int32, "label")]
    feature_descr = [(f"f{i}", float) for i in range(83)]

    ret = _fetch_remote_csv(ARCHIVE, file_name,
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=85, all_num=True
                            )
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""Nested models where control outcome predictions are used in an
uplift model."""

import numpy as np

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression

from ..utils import safe_hstack

from .base import UpliftMetaModelBase
from ..base import UpliftRegressorMixin
from ..base import UpliftClassifierMixin

class NestedMeanUpliftRegressor(UpliftRegressorMixin, UpliftMetaModelBase):
    """Nested regression model.

    First builds a model on controls, then subtracts its training
    predictions from target.  An uplift model is then build on the new
    target.

    Only available for regression models.

    """
    def __init__(self, base_estimator = LinearRegression()):
        super().__init__(base_estimator=base_estimator)
    def _get_model_names_list(self, X=None, y=None, trt=None):
        m_names = ["model_c"]
        for i in range(self.n_trt_):
            name = "model_u"
            if self.n_trt_ > 1:
                name += str(i-1)
            m_names.append(name)
        return m_names
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        c_mask = (trt==0)
        y_c = y[c_mask]
        if sample_weight is not None:
            yield X[c_mask], y_c, sample_weight[c_mask]
        else:
            yield X[c_mask], y_c, None
        # assume the control model is already fitted
        m_c = self.models_[0][1]
        for i in range(self.n_trt_):
            t_mask = (trt==(i+1))
            X_i = X[t_mask]
            y_i_pred = m_c.predict(X_i)
            y_i = y[t_mask] - y_i_pred
            if sample_weight is not None:
                w_i = sample_weight[t_mask]
            else:
                w_i = None
            yield X_i, y_i, w_i
    def predict(self, X):
        preds = [m_i.predict(X) for _, m_i in self.models_[1:]]
        if self.n_trt_ == 1:
            y = preds[0]
        else:
            y = np.column_stack(preds)
        return y


class DDRUpliftClassifier(UpliftClassifierMixin, UpliftMetaModelBase):
    """Dependent Data Representation metamodel.  It is a double model
    where control predictions are added as a variable in the treatment
    model.

    The model was proposed in A. Betlei, E. Diemert, and M.-R. Amini
    Uplift Prediction with Dependent Feature Representation in
    Imbalanced Treatment and Control Conditions, ICONIP, 2018.

    direction : string, default="C->T" "C->T" means control
        predictions are used as an additional predictor for the
        treatment model, "T->C" means the reverse: predictions of all
        treatment models are used (jointly) as predictors for the
        control model.

    """
    def __init__(self, base_estimator=LogisticRegression(),
                 feature_prediction_method="predict_proba",
                 direction="C->T"):
        """A DDR uplift """
        super().__init__(base_estimator=base_estimator)
        self.feature_prediction_method = feature_prediction_method
        self.direction = direction
    def _get_model_names_list(self, X=None, y=None, trt=None):
        c_names = ["model_c"]
        t_names = []
        for i in range(self.n_trt_):
            name = "model_t"
            if self.n_trt_ > 1:
                name += str(i-1)
            t_names.append(name)
        if self.direction == "C->T":
            m_names = c_names + t_names
        elif self.direction == "T->C":
            m_names = t_names + c_names
        else:
            raise ValueError(f"The direction parameter for the DDR model must be 'C->T' or 'T->C', got {self.direction}")
        return m_names
    def _prediction_feature(self, m, X):
        if self.feature_prediction_method == "predict_proba":
            y = m.predict_proba(X)[:,1:]
        elif self.feature_prediction_method == "predict":
            y = m.predict(X)
        elif self.feature_prediction_method == "decision_function":
            y = m.decision_function(X)
        return y
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        c_mask = (trt==0)
        if self.direction == "C->T":
            if sample_weight is not None:
                yield X[c_mask], y[c_mask], sample_weight[c_mask]
            else:
                yield X[c_mask], y[c_mask], None
            m_c = self.models_[0][1]
            for i in range(self.n_trt_):
                t_mask = (trt==(i+1))
                X_i = X[t_mask]
                y_i = y[t_mask]
                # assume the control model is already fitted
                y_i_pred = self._prediction_feature(m_c, X_i)
                X_i = safe_hstack([X_i, y_i_pred])
                if sample_weight is not None:
                    w_i = sample_weight[t_mask]
                else:
                    w_i = None
                yield X_i, y_i, w_i
        else:
            X_c = X[c_mask]
            # train treatment models first
            t_preds = []
            for i in range(self.n_trt_):
                t_mask = (trt==(i+1))
                X_i = X[t_mask]
                y_i = y[t_mask]
                # assume the control model is already fitted
                if sample_weight is not None:
                    w_i = sample_weight[t_mask]
                else:
                    w_i = None
                yield X_i, y_i, w_i
                y_i_pred = self._prediction_feature(self.models_[i][1], X_c)
                t_preds.append(y_i_pred)
            # train the control model
            X_c = safe_hstack([X_c] + t_preds)
            y_c = y[c_mask]
            if sample_weight is not None:
                w_c = sample_weight[c_mask]
            else:
                w_c = None
            yield X_c, y_c, w_c
            
    def predict(self, X):
        if self.direction == "C->T":
            pred_feature = self._prediction_feature(self.models_[0][1], X)
            pred_c = self.models_[0][1].predict_proba(X)
            preds = []
            for i in range(self.n_trt_):
                X_i = safe_hstack([X, pred_feature])
                pred_i = self.models_[i+1][1].predict_proba(X_i) - pred_c
                preds.append(pred_i)
        else:
            pred_features = [self._prediction_feature(m[1], X) for m in self.models_[:-1]]
            preds = [m[1].predict_proba(X) for m in self.models_[:-1]]
            X_c = safe_hstack([X] + pred_features)
            pred_c = self.models_[-1][1].predict_proba(X_c)
            for i in range(self.n_trt_):
                preds[i] -= pred_c
        if self.n_trt_ == 1:
            y = preds[0]
        else:
            y = np.dstack(preds)
        return y



"""Measures based on comparing treatment and control statistics within
bins, such as quantiles."""

import numpy as np

from sklearn.utils.validation import check_array, check_consistent_length

from ..utils.validation import check_trt
from ..utils.stats import quantile

def iter_quantiles(scores, trt, n_trt, n=10, joint=False,
                   sample_weight=None):
    """Iterate simultaneously over quantiles of score vectors for all
    treatments.

    Returns a generator which, for each quantile, returns a list of
    index arrays for scores within each treatment.

    If joint is True, quantiles are computed jointly for all
    treatments.

    If sample_weight is not None, weighted quantiles are used.

    """
    # sort by treatment
    t_idx = np.argsort(trt)
    t_counts = np.r_[[0], np.cumsum(np.bincount(trt, minlength=n_trt+1))]
    counts = []
    idxs = []
    if joint:
        q = quantile(scores, np.linspace(0,1,n, endpoint=False),
                     weights=sample_weight)
    for ti in range(n_trt+1):
        s = scores[t_idx[t_counts[ti]:t_counts[ti+1]]]
        if not joint:
            if sample_weight is not None:
                w = sample_weight[t_idx[t_counts[ti]:t_counts[ti+1]]]
            else:
                w = None
            q = quantile(s, np.linspace(0,1,n, endpoint=False),
                         weights=w)
        b = (np.digitize(s, q)-1).astype(np.min_scalar_type(n+1))
        c = np.r_[[0], np.cumsum(np.bincount(b, minlength=n))]
        idx = np.argsort(b)
        counts.append(c)
        idxs.append(t_idx[t_counts[ti]:t_counts[ti+1]][idx])
    for i in range(n-1, -1, -1):
        yield [idx[c[i]:c[i+1]] for idx, c in zip(idxs, counts)]

def _binned_measure(per_q_func, aggreg_func, name,
                    y_true, y_pred, trt, n_trt=None, sample_weight=None,
                    n_bins=10, allow_nans=False, joint_quantiles=False):
    y_true = check_array(y_true, ensure_2d=False)
    y_pred = check_array(y_pred, ensure_2d=False)
    trt, n_trt = check_trt(trt, n_trt)
    if sample_weight is None:
        check_consistent_length(y_true, y_pred, trt)
    else:
        sample_weight = check_array(sample_weight, ensure_2d=False)
        check_consistent_length(y_true, y_pred, trt, sample_weight)
    if n_trt > 1:
        raise ValueError(name + " only supported for a single treatment.")

    stat_q_s = []
    for q_idxs in iter_quantiles(y_pred, trt, n_trt, n_bins, joint=joint_quantiles):
        idx_c, idx_t = q_idxs
        stat_q = per_q_func(y_true, y_pred, idx_t, idx_c)
        stat_q_s.append(stat_q)
    stat = aggreg_func(stat_q_s, allow_nans)
    if np.isnan(stat):
        raise RuntimeError("NaN's present in " + name + " computation")
    return stat

def _q_aggreg_mean(stats_q_s, allow_nans):
    if allow_nans:
        stat = np.nanmean(stats_q_s)
    else:
        stat = np.mean(stats_q_s)
    return stat
def _q_aggreg_max(stats_q_s, allow_nans):
    if allow_nans:
        stat = np.nanmax(stats_q_s)
    else:
        stat = np.max(stats_q_s)
    return stat

def QMSE(y_true, y_pred, trt, n_trt=None, sample_weight=None,
         n_bins=10, allow_nans=False, joint_quantiles=False):
    """The per-quantile MSE measure by Rudaś, Jaroszewicz."""
    def _per_q_qmse(y_true, y_pred, idx_t, idx_c):
        if len(idx_c) > 0:
            mi_c = y_true[idx_c].mean()
        else:
            return np.nan
        if len(idx_t) > 0:
            mi_t = y_true[idx_t].mean()
        else:
            return np.nan
        if len(idx_t) > 0:
            mse_j = np.mean((y_pred[idx_t] - (mi_t-mi_c))**2)
        else:
            mse_j = np.nan
        return mse_j
    return _binned_measure(_per_q_qmse, _q_aggreg_mean, "QMSE",
                           y_true, y_pred, trt, n_trt=n_trt, sample_weight=sample_weight,
                           n_bins=10, allow_nans=allow_nans, joint_quantiles=joint_quantiles)
def QMSE_j(y_true, y_pred, trt, n_trt=None, sample_weight=None,
         n_bins=10, allow_nans=False):
    return QMSE(y_true, y_pred, trt, n_trt=None, sample_weight=None,
                n_bins=10, allow_nans=False, joint_quantiles=True)

def _per_q_euce(y_true, y_pred, idx_t, idx_c):
    if len(idx_c) > 0:
        mi_c = y_true[idx_c].mean()
    else:
        return np.nan
    if len(idx_t) > 0:
        mi_t = y_true[idx_t].mean()
    else:
        return np.nan
    idx = np.concatenate([idx_c, idx_t])
    mi_p = np.mean(y_pred[idx])
    d_j = np.abs(mi_p - (mi_t-mi_c))
    return d_j

def EUCE(y_true, y_pred, trt, n_trt=None, sample_weight=None,
         n_bins=100, allow_nans=False, joint_quantiles=True):
    """The EUCE measure by Nyberg and Klami."""
    return _binned_measure(_per_q_euce, _q_aggreg_mean, "EUCE",
                           y_true, y_pred, trt, n_trt=n_trt, sample_weight=sample_weight,
                           n_bins=10, allow_nans=allow_nans, joint_quantiles=joint_quantiles)
def MUCE(y_true, y_pred, trt, n_trt=None, sample_weight=None,
         n_bins=100, allow_nans=False, joint_quantiles=True):
    """The MUCE measure by Nyberg and Klami."""
    return _binned_measure(_per_q_euce, _q_aggreg_max, "MUCE",
                           y_true, y_pred, trt, n_trt=n_trt, sample_weight=sample_weight,
                           n_bins=10, allow_nans=allow_nans, joint_quantiles=joint_quantiles)



"""A memoized classifier class.

Used to avoid recomputing the same classifier twice e.g. when both
T-learner and Response models are computed.

"""

import os
from tempfile import gettempdir

from sklearn.base import BaseEstimator
from sklearn.base import clone
from sklearn.utils.validation import check_memory

def _do_fit(estimator, model_params, *args, **kwargs):
    return clone(estimator).fit(*args, **kwargs)

class MemoizedClassifier(BaseEstimator):
    def __init__(self, estimator, memory=None):
        """Creates a memoized version of estimator.

        Subsequent calls to fit with the same arguments will reuse a
        prefitted model.

        memory is either a path or a joblib.Memory object.  If None a
        default path is used: "usklearn_cache" in systems default
        temporary directory.

        """
        self.estimator = estimator
        self.memory = memory
    def fit(self, *args, **kwargs):
        memory = self.memory
        if memory is None:
            memory = os.path.join(gettempdir(), "usklearn_cache")
        self.memory_ = check_memory(memory)
        if not hasattr(self, "do_fit_cached_"):
            self.do_fit_cached_ = self.memory_.cache(_do_fit, ignore=["estimator"])
        self.fitted_estimator_ = self.do_fit_cached_(self.estimator,
                                                     self.estimator.get_params(),
                                                     *args, **kwargs)
        return self
    
    def __getattr__(self, name):
        if name in ["fitted_estimator_", "do_fit_cached_"]:
            try:
                return self.__dict__[name]
            except:
                raise AttributeError(f"MemoizedClassifier has no attribute {name}")
        if "fitted_estimator_" not in self.__dict__:
            return getattr(self.estimator, name)
        return getattr(self.fitted_estimator_, name)




from .regression import e_sate, e_satt
from .curves import uplift_curve, uplift_curve_j
from .curves import area_under_uplift_curve, area_under_uplift_curve_j
from .bins import iter_quantiles
from .bins import QMSE, QMSE_j, EUCE, MUCE
from ._scorer import (
    make_uplift_scorer,
    check_uplift_scoring,
    get_uplift_scorer,
    get_uplift_scorer_names
)



"""Uplift and Qini curves."""

import numpy as np

from sklearn.utils.validation import check_array, check_consistent_length

from ..utils import check_trt
from ..utils import area_under_curve

def _cumulative_gains_curve(y_true, y_score, sample_weight):
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    if sample_weight is not None:
        weight = sample_weight[desc_score_indices]
    else:
        weight = 1.0

    # handle tied values
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    # compute gains and prepend (0,0) point at the beginning
    gains = np.r_[0, np.cumsum(y_true * weight)[threshold_idxs]]
    if sample_weight is not None:
        xs = np.r_[0, np.cumsum(weight)[threshold_idxs]]
        xs = xs / xs[-1]
    else:
        xs = np.r_[0, threshold_idxs+1]
        xs = xs / xs[-1]
    return xs, gains

def uplift_curve(y_true, y_score, trt, n_trt=None, pos_label=None,
                 sample_weight=None):
    """Uplift curve.

    Unless specified explicitly, y_true is assumed to be 0-1, with 1
    the positive outcome.

    This function implements the variant used by Rzepakowski and
    Jaroszewicz, where treatment and control curves are computed
    separately and subtracted.

    """
    y_true = check_array(y_true, ensure_2d=False)
    y_score = check_array(y_score, ensure_2d=False)
    trt, n_trt = check_trt(trt, n_trt)
    if sample_weight is None:
        check_consistent_length(y_true, y_score, trt)
        sample_weight_c = None
        sample_weight_t = None
        n_c = (trt==0).sum()
        n_t = (trt==1).sum()
    else:
        sample_weight = check_array(sample_weight, ensure_2d=False)
        check_consistent_length(y_true, y_score, trt, sample_weight)
        sample_weight_c = sample_weight[trt==0]
        sample_weight_t = sample_weight[trt==1]
        n_c = sample_weight_c.sum()
        n_t = sample_weight_t.sum()
    if n_trt > 1:
        raise ValueError("uplift curve only supported for a single treatment.")
        
    if pos_label is not None:
        y_true = (y_true == pos_label)

    y_score_c = y_score[trt==0]
    y_score_t = y_score[trt==1]
    y_true_c = y_true[trt==0]
    y_true_t = y_true[trt==1]
    
    x_c, gains_c = _cumulative_gains_curve(y_true_c, y_score_c, sample_weight_c)
    x_t, gains_t = _cumulative_gains_curve(y_true_t, y_score_t, sample_weight_t)

    # normalize
    if n_c == 0:
        raise RuntimeError("Cannot construct uplift curve: no cases in control")
    if n_t == 0:
        raise RuntimeError("Cannot construct uplift curve: no treated cases")
    gains_c /= n_c
    gains_t /= n_t

    # interpolate and subtract curves
    x = np.union1d(x_c, x_t)
    y_c = np.interp(x, x_c, gains_c)
    y_t = np.interp(x, x_t, gains_t)
    u = y_t - y_c
    return x, u

def uplift_curve_j(y_true, y_score, trt, n_trt=None, pos_label=None, sample_weight=None):
    """Uplift curve.

    Unless specified explicitly, y_true is assumed to be 0-1, with 1
    the positive outcome.

    This function implements the variant where scores are sorted
    jointly, see Verbeke, Nyberg, Verhelst.

    """
    y_true = check_array(y_true, ensure_2d=False, copy=True, dtype=float)
    y_score = check_array(y_score, ensure_2d=False)
    trt, n_trt = check_trt(trt, n_trt)
    if sample_weight is None:
        check_consistent_length(y_true, y_score, trt)
        sample_weight = np.ones_like(y_true, dtype=float)
    else:
        sample_weight = check_array(sample_weight, ensure_2d=False,
                                    dtype=float, copy=True)
        check_consistent_length(y_true, y_score, trt, sample_weight)
    if n_trt > 1:
        raise ValueError("uplift curve only supported for a single treatment.")
        
    if pos_label is not None:
        y_true = (y_true == pos_label)

    # normalize weights
    n_c = sample_weight[trt==0].sum()
    n_t = sample_weight[trt==1].sum()
    if n_c == 0:
        raise RuntimeError("Cannot construct uplift curve: no cases in control")
    if n_t == 0:
        raise RuntimeError("Cannot construct uplift curve: no treated cases")
    y_true[trt==0] = -y_true[trt==0]
    sample_weight[trt==0] /= n_c
    sample_weight[trt==1] /= n_t
    
    x, u = _cumulative_gains_curve(y_true, y_score, sample_weight)

    return x, u


def area_under_uplift_curve(y_true, y_score, trt, n_trt=None, pos_label=None, sample_weight=None,
         subtract_diag=True):
    x, u = uplift_curve(y_true, y_score, trt, n_trt=n_trt, pos_label=pos_label,
                        sample_weight=sample_weight)
    return area_under_curve(x, u, subtract_diag=subtract_diag)
def area_under_uplift_curve_j(y_true, y_score, trt, n_trt=None, pos_label=None, sample_weight=None,
           subtract_diag=True):
    x, u = uplift_curve_j(y_true, y_score, trt, n_trt=n_trt, pos_label=pos_label,
                          sample_weight=sample_weight)
    return area_under_curve(x, u, subtract_diag=subtract_diag)



"""Multiple simultaneuosly indexed arrays.

Needed to work around lack of sample properties in scikit-learn."""

import numpy as np

from sklearn.utils import check_consistent_length

class MultiArray:
    __array_ufunc__ = None # don't allow numpy operations by default

    def __init__(self, main_array, array_dict=None, scalar_dict=None):
        """behaves like main_array w.r.t. indexing, but arrays in

        array_dict are indexed simulteneously, scalar_dict is passed to
        indexing results."""
        self.main_array = main_array
        if array_dict is None:
            array_dict = dict()
        check_consistent_length([main_array] + list(array_dict.values()))
        self.array_dict = array_dict
        if scalar_dict is None:
            scalar_dict = dict()
        self.scalar_dict = scalar_dict
        self.shape = self.main_array.shape
        self.ndim = self.main_array.ndim
        self.dtype = self.main_array.dtype
    def __getitem__(self, idx):
        new_dict = {k:self.array_dict[k][idx] for k in self.array_dict}
        return MultiArray(self.main_array[idx], new_dict, self.scalar_dict)



"""Implement scorers for uplift modeling."""

import copy

from sklearn.base import is_classifier

from .regression import e_sate, e_satt
from .curves import area_under_uplift_curve, area_under_uplift_curve_j
from .bins import QMSE, QMSE_j, EUCE, MUCE

class _BaseUpliftScorer:
    def __init__(self, score_func, sign, kwargs):
        self._kwargs = kwargs
        self._score_func = score_func
        self._sign = sign

class _UpliftPredictScorer(_BaseUpliftScorer):
    def __call__(self, estimator, X, y_true, trt, n_trt=None, sample_weight=None):
        """Evaluate predicted target values for X relative to y_true.

        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.

        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.

        X : array-like or sparse matrix
            Test data that will be fed to estimator.predict.

        y_true : array-like
            Gold standard target values for X.

        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        y_pred = estimator.predict(X)
        if is_classifier(estimator):
            if "pos_label" in self._kwargs:
                pos_label = self._kwargs["pos_label"]
            else:
                pos_label = 1
            y_pred = y_pred[:,pos_label]
        if sample_weight is not None:
            return self._sign * self._score_func(y_true, y_pred, trt, n_trt,
                                                 sample_weight=sample_weight,
                                                 **self._kwargs)
        else:
            return self._sign * self._score_func(y_true, y_pred, trt, n_trt,
                                                 **self._kwargs)
class _UpliftDecisionScorer(_BaseUpliftScorer):
    def _score(self, estimator, X, y_true, trt, n_trt=None, sample_weight=None):
        """Evaluate predicted treatment decisions for X relative to y_true.

        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.

        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.

        X : array-like or sparse matrix
            Test data that will be fed to estimator.predict.

        y_true : array-like
            Gold standard target values for X.

        sample_weight : array-like, optional (default=None)
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """
        a = estimator.predict_action(X)
        if sample_weight is not None:
            return self._sign * self._score_func(y_true, a, trt, n_trt,
                                                 sample_weight=sample_weight,
                                                 **self._kwargs)
        else:
            return self._sign * self._score_func(y_true, a, trt, n_trt,
                                                 **self._kwargs)
class _PassthroughUpliftScorer:
    def __init__(self, estimator):
        self._estimator = estimator

    def __call__(self, estimator, *args, **kwargs):
        """Method that wraps estimator.score"""
        return estimator.score(*args, **kwargs)

    #def get_metadata_routing(self):
    #    """Get requested data properties.
    #    .. versionadded:: 1.2
    #    Returns
    #    -------
    #    routing : MetadataRouter
    #        A :class:`~utils.metadata_routing.MetadataRouter` encapsulating
    #        routing information.
    #    """
    #    # This scorer doesn't do any validation or routing, it only exposes the
    #    # score requests to the parent object. This object behaves as a
    #    # consumer rather than a router.
    #    res = MetadataRequest(owner=self._estimator.__class__.__name__)
    #    res.score = get_routing_for_object(self._estimator).score
    #    return res

def get_uplift_scorer(scoring):
    """Get an uplift scorer from string.
    Parameters
    ----------
    scoring : str or callable
        Scoring method as string. If callable it is returned as is.
    Returns
    -------
    scorer : callable
        The scorer.
    Notes
    -----
    When passed a string, this function always returns a copy of the scorer
    object. Calling `get_uplift_scorer` twice for the same scorer results in two
    separate scorer objects.
    """
    if isinstance(scoring, str):
        try:
            scorer = copy.deepcopy(_UPLIFT_SCORERS[scoring])
        except KeyError:
            raise ValueError(
                "%r is not a valid uplift scoring value. "
                "Use usklearn.metrics.get_uplift_scorer_names() "
                "to get valid options." % scoring
            )
    else:
        scorer = scoring
    return scorer

def check_uplift_scoring(estimator, scoring=None, *, allow_none=False):
    """Determine uplift scorer from user options.
    A TypeError will be thrown if the estimator cannot be scored.
    Parameters
    ----------
    estimator : estimator object implementing 'fit'
        The object to use to fit the data.
    scoring : str or callable, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        ``scorer(estimator, X, y, trt, n_trt)``.
        If None, the provided estimator object's `score` method is used.
    allow_none : bool, default=False
        If no scoring is specified and the estimator has no score function, we
        can either return None or raise an exception.
    Returns
    -------
    scoring : callable
        A scorer callable object / function with signature
        ``scorer(estimator, X, y)``.
    """
    if not hasattr(estimator, "fit"):
        raise TypeError(
            "estimator should be an estimator implementing 'fit' method, %r was passed"
            % estimator
        )
    if isinstance(scoring, str):
        return get_uplift_scorer(scoring)
    elif callable(scoring):
        # Heuristic to ensure user has not passed a metric
        module = getattr(scoring, "__module__", None)
        if (
            hasattr(module, "startswith")
            and module.startswith("usklearn.metrics.")
            and not module.startswith("usklearn.metrics._scorer")
            and not module.startswith("usklearn.metrics.tests.")
        ):
            raise ValueError(
                "scoring value %r looks like it is a metric "
                "function rather than a scorer. A scorer should "
                "require an estimator as its first parameter. "
                "Please use `make_scorer` to convert a metric "
                "to a scorer." % scoring
            )
        return get_uplift_scorer(scoring)
    elif scoring is None:
        if hasattr(estimator, "score"):
            return _PassthroughUpliftScorer(estimator)
        elif allow_none:
            return None
        else:
            raise TypeError(
                "If no scoring is specified, the estimator passed should "
                "have a 'score' method. The estimator %r does not." % estimator
            )
    elif isinstance(scoring, Iterable):
        raise ValueError(
            "For evaluating multiple scores, use "
            "usklearn.model_selection.cross_validate instead. "
            "{0} was passed.".format(scoring)
        )
    else:
        raise ValueError(
            "uplift scoring value should either be a callable, string or None. %r was passed"
            % scoring
        )



def make_uplift_scorer(score_func, greater_is_better=True, needs_decision=False,
                       needs_proba=False, needs_threshold=False, **kwargs):
    """Make a scorer from a performance metric or loss function.
    This factory function wraps scoring functions for use in GridSearchCV
    and cross_val_score. It takes a score function, such as ``accuracy_score``,
    ``mean_squared_error``, ``adjusted_rand_index`` or ``average_precision``
    and returns a callable that scores an estimator's output.
    Read more in the :ref:`User Guide <scoring>`.
    Parameters
    ----------
    score_func : callable,
        Score function (or loss function) with signature
        ``score_func(y, y_pred, **kwargs)``.
    greater_is_better : boolean, default=True
        Whether score_func is a score function (default), meaning high is good,
        or a loss function, meaning low is good. In the latter case, the
        scorer object will sign-flip the outcome of the score_func.
    needs_proba : boolean, default=False
        Whether score_func requires predict_proba to get probability estimates
        out of a classifier.
    needs_threshold : boolean, default=False
        Whether score_func takes a continuous decision certainty.
        This only works for binary classification using estimators that
        have either a decision_function or predict_proba method.
        For example ``average_precision`` or the area under the roc curve
        can not be computed using discrete predictions alone.
    **kwargs : additional arguments
        Additional parameters to be passed to score_func.
    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score; greater is better.
    Examples
    --------
    >>> from sklearn.metrics import fbeta_score, make_scorer
    >>> ftwo_scorer = make_scorer(fbeta_score, beta=2)
    >>> ftwo_scorer
    make_scorer(fbeta_score, beta=2)
    >>> from sklearn.model_selection import GridSearchCV
    >>> from sklearn.svm import LinearSVC
    >>> grid = GridSearchCV(LinearSVC(), param_grid={'C': [1, 10]},
    ...                     scoring=ftwo_scorer)
    """
    if needs_proba or needs_threshold:
        raise NotImplementedError()
    sign = 1 if greater_is_better else -1
    if needs_proba + needs_decision + needs_threshold > 1:
        raise ValueError("Set only one of needs_proba, needs_threshold or"
                         " needs_decision to True.")
    if needs_proba:
        cls = _UpliftProbaScorer
    elif needs_threshold:
        cls = _UpliftThresholdScorer
    elif needs_decision:
        cls = _UpliftDecisionScorer
    else:
        cls = _UpliftPredictScorer
    return cls(score_func, sign, kwargs)


_UPLIFT_SCORERS = dict(
    e_sate=make_uplift_scorer(e_sate, greater_is_better=False),
    e_satt=make_uplift_scorer(e_satt, greater_is_better=False),
    auuc=make_uplift_scorer(area_under_uplift_curve, greater_is_better=True),
    auuc_j=make_uplift_scorer(area_under_uplift_curve_j, greater_is_better=True),
    QMSE=make_uplift_scorer(QMSE, greater_is_better=False),
    QMSE_j=make_uplift_scorer(QMSE_j, greater_is_better=False),
    EUCE=make_uplift_scorer(EUCE, greater_is_better=False),
    MUCE=make_uplift_scorer(MUCE, greater_is_better=False),
)

def get_uplift_scorer_names():
    """Get the names of all available scorers.
    These names can be passed to :func:`~usklearn.metrics.get_uplift_scorer` to
    retrieve the scorer object.
    Returns
    -------
    list of str
        Names of all available scorers.
    """
    return sorted(_UPLIFT_SCORERS.keys())



"""The marketing campaign dataset from the CRAN Information package by
Kim Larsen.

See: https://cran.r-project.org/web/packages/Information/index.html
for details.

"""

import numpy as np

from .base import _fetch_remote_csv
from .base import RemoteFileMetadata


ARCHIVE_T = RemoteFileMetadata(
    filename="Information_train.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Information/Information_train.csv.gz'),
    checksum=('7632536786357871de3f2438dbbe6c70'
              '17829b2d244b2878dba9fb11af0a449b'))
ARCHIVE_V = RemoteFileMetadata(
    filename="Information_valid.csv.gz",
    url=('https://github.com/jszymon/uplift_sklearn_data/'
         'releases/download/Information/Information_valid.csv.gz'),
    checksum=('96b1773bc9fea564de03d412b936b32a'
              'b5d3b1a9f7338dbb59aa6f918806fe66'))


def fetch_Information(version="train", data_home=None,
                      download_if_missing=True,
                      random_state=None, shuffle=False,
                      categ_as_strings=False, return_X_y=False,
                      as_frame=False):
    """Load the marketing campaign dataset from the CRAN Information
    package by Kim Larsen.

    See: https://cran.r-project.org/web/packages/Information/index.html

    Two datasets are available: "train" and "validation".  Use version
    argument to select.
    
    Download it if necessary.

    Parameters
    ----------
    data_home : string, optional
        Specify another download and cache folder for the datasets. By default
        all scikit-learn data is stored in '~/scikit_learn_data' subfolders.

    download_if_missing : boolean, default=True
        If False, raise a IOError if the data is not locally available
        instead of trying to download the data from the source site.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for dataset shuffling. Pass an int
        for reproducible output across multiple function calls.

    shuffle : bool, default=False
        Whether to shuffle dataset.

    categ_as_strings : bool, default=False
        Whether to return categorical variables as strings.

    return_X_y : boolean, default=False.
        If True, returns ``(data.data, data.target)`` instead of a Bunch
        object.

    as_frame : boolean, default=False
        If True features are returned as pandas DataFrame.  If False
        features are returned as object or float array.  Float array
        is returned if all features are floats.
    
    Returns
    -------
    dataset : dict-like object with the following attributes:

    dataset.data : numpy array
        Each row corresponds to the features in the dataset.

    dataset.target : numpy array
        Each value is 1 if a purchase was made 0 otherwise.

    dataset.DESCR : string
        Description of the dataset.

    (data, target_conversion, target_visit, target_exposure) : tuple if
        ``return_X_y`` is True

    """

    # dictionaries
    treatment_values = ['control', 'treated']
    categ_values = dict()

    # attribute descriptions
    treatment_descr = [("treatment", np.int32, "TREATMENT")]
    target_descr = [("target", np.int32, "PURCHASE"),
                    ]
    feature_descr = [('AGE', float),
                     ('AGRGT_BAL_ALL_XCLD_MRTG', float),
                     ('AUTO_2_OPEN_DATE_YRS', float),
                     ('AUTO_HI_CRDT_2_ACTUAL', float),
                     ('AVG_BAL_ALL_FNC_REV_ACTS', float),
                     ('AVG_BAL_ALL_PRM_BC_ACTS', float),
                     ('D_DEPTCARD', float),
                     ('D_NA_AVG_BAL_ALL_FNC_REV_ACTS', float),
                     ('D_NA_M_SNCOLDST_BNKINSTL_ACTOPN', float),
                     ('D_NA_M_SNC_MST_RCNT_ACT_OPN', float),
                     ('D_NA_M_SNC_MST_RCNT_MRTG_DEAL', float),
                     ('D_NA_M_SNC_OLDST_MRTG_ACT_OPN', float),
                     ('D_NA_RATIO_PRSNL_FNC_BAL2HICRDT', float),
                     ('D_REGION_A', float),
                     ('D_REGION_B', float),
                     ('D_REGION_C', float),
                     ('FNC_CARD_OPEN_DATE_YRS', float),
                     ('HI_RETAIL_CRDT_LMT', float),
                     ('MAX_MRTG_CLOSE_DATE', float),
                     ('MRTG_1_MONTHLY_PAYMENT', float),
                     ('MRTG_2_CURRENT_BAL', float),
                     ('M_SNCOLDST_BNKINSTL_ACTOPN', float),
                     ('M_SNCOLDST_OIL_NTN_TRD_OPN', float),
                     ('M_SNC_MSTRCNT_MRTG_ACT_UPD', float),
                     ('M_SNC_MSTREC_INSTL_TRD_OPN', float),
                     ('M_SNC_MST_RCNT_60_DAY_RTNG', float),
                     ('M_SNC_MST_RCNT_ACT_OPN', float),
                     ('M_SNC_MST_RCNT_MRTG_DEAL', float),
                     ('M_SNC_OLDST_MRTG_ACT_OPN', float),
                     ('M_SNC_OLDST_RETAIL_ACT_OPN', float),
                     ('N30D_ORWRS_RTNG_MRTG_ACTS', float),
                     ('N_120D_RATINGS', float),
                     ('N_30D_AND_60D_RATINGS', float),
                     ('N_30D_RATINGS', float),
                     ('N_ACTS_90D_PLS_LTE_IN_6M', float),
                     ('N_ACTS_WITH_MXD_3_IN_24M', float),
                     ('N_ACTS_WITH_MXD_4_IN_24M', float),
                     ('N_BANK_INSTLACTS', float),
                     ('N_BC_ACTS_OPN_IN_12M', float),
                     ('N_BC_ACTS_OPN_IN_24M', float),
                     ('N_DEROG_PUB_RECS', float),
                     ('N_DISPUTED_ACTS', float),
                     ('N_FNC_ACTS_OPN_IN_12M', float),
                     ('N_FNC_ACTS_VRFY_IN_12M', float),
                     ('N_FNC_INSTLACTS', float),
                     ('N_INQUIRIES', float),
                     ('N_OF_MRTG_ACTS_DLINQ_24M', float),
                     ('N_OF_SATISFY_FNC_REV_ACTS', float),
                     ('N_OPEN_REV_ACTS', float),
                     ('N_PUB_REC_ACT_LINE_DEROGS', float),
                     ('N_RETAIL_ACTS_OPN_IN_24M', float),
                     ('N_SATISFY_INSTL_ACTS', float),
                     ('N_SATISFY_OIL_NATIONL_ACTS', float),
                     ('N_SATISFY_PRSNL_FNC_ACTS', float),
                     ('PRCNT_OF_ACTS_NEVER_DLQNT', float),
                     ('PREM_BANKCARD_CRED_LMT', float),
                     ('RATIO_BAL_TO_HI_CRDT', float),
                     ('RATIO_PRSNL_FNC_BAL2HICRDT', float),
                     ('RATIO_RETAIL_BAL2HI_CRDT', float),
                     ('STUDENT_HI_CRED_RANGE', float),
                     ('STUDENT_OPEN_DATE_YRS', float),
                     ('TOT_BAL_ALL_DPT_STORE_ACTS', float),
                     ('TOT_HI_CRDT_CRDT_LMT', float),
                     ('TOT_INSTL_HI_CRDT_CRDT_LMT', float),
                     ('TOT_NOW_LTE', float),
                     ('TOT_OTHRFIN_HICRDT_CRDTLMT', float),
                     ('UPSCALE_OPEN_DATE_YRS', float),
                     ]


    if version == "train":
        arch = ARCHIVE_T
    elif version == "valid":
        arch = ARCHIVE_V
    else:
        raise ValueError("Wrong Information dataset version."
                         "  Got '{version}', expected 'train' or 'valid'.")
    ret = _fetch_remote_csv(arch, "Information",
                            feature_attrs=feature_descr,
                            treatment_attrs=treatment_descr,
                            target_attrs=target_descr,
                            categ_as_strings=categ_as_strings,
                            return_X_y=return_X_y, as_frame=as_frame,
                            download_if_missing=download_if_missing,
                            random_state=random_state, shuffle=shuffle,
                            total_attrs=69, all_num=True
                            )
    if not return_X_y:
        ret.descr = __doc__
    return ret



"""The S-learner meta model.

Simply add the treatment variable to a classifier/regressor.
"""

import numpy as np

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression

from ..utils import safe_hstack

from .base import UpliftMetaModelBase
from ..base import UpliftRegressorMixin
from ..base import UpliftClassifierMixin



class _SLearnerBase(UpliftMetaModelBase):
    def __init__(self, base_estimator, treatment_encoding="one_hot"):
        super().__init__(base_estimator=base_estimator)
        self.treatment_encoding = treatment_encoding
    def _get_model_names_list(self, X=None, y=None, trt=None):
        m_names = ["model"]
        return m_names
    def _encode_treatment(self, trt):
        if self.n_trt_ == 1 or self.treatment_encoding == "int":
            trt = trt.reshape(-1, 1).astype(float)
        elif self.treatment_encoding == "one_hot":
            trt = np.eye(self.n_trt_+1)[:,1:]
        else:
            raise ValueError(f"Unsupported threatment encoding"
                             "({self.treatment_encoding}) for SLearner")
        return trt
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        trt = self._encode_treatment(trt)
        X = safe_hstack([X, trt])
        yield X, y, sample_weight
    def _predict_diffs(self, X, prediction_method):
        n = X.shape[0]
        p = X.shape[1]
        trt_0 = self._encode_treatment(np.zeros(n, dtype=int))
        X_aug = safe_hstack([X, trt_0])
        y_0 = getattr(self.models_[0][1], prediction_method)(X_aug)
        pred_diffs = []
        for i in range(self.n_trt_):
            trt_i = self._encode_treatment(np.full(n, i+1, dtype=int))
            X_aug[:,p:] = trt_i
            y_i = getattr(self.models_[0][1], prediction_method)(X_aug)
            pred_diffs.append(y_i - y_0)
        return pred_diffs

class SLearnerUpliftRegressor(UpliftRegressorMixin, _SLearnerBase):
    def __init__(self, base_estimator=LinearRegression(),
                 treatment_encoding="one_hot"):
        """The S-learner meta regressor.

        treatment can be encoded either as integer or using one-hot
        encoding (with control indicator skipped).

        """
        super().__init__(base_estimator=base_estimator,
                         treatment_encoding=treatment_encoding)
    def predict(self, X):
        pred_diffs = self._predict_diffs(X, "predict")
        if self.n_trt_ == 1:
            y = pred_diffs[0]
        else:
            y = np.column_stack(pred_diffs)
        return y

class SLearnerUpliftClassifier(UpliftClassifierMixin, _SLearnerBase):
    def __init__(self, base_estimator=LogisticRegression(),
                 treatment_encoding="one_hot"):
        """The S-learner meta regressor.

        treatment can be encoded either as integer or using one-hot
        encoding (with control indicator skipped).

        """
        super().__init__(base_estimator=base_estimator,
                         treatment_encoding=treatment_encoding)
    def predict(self, X):
        pred_diffs = self._predict_diffs(X, "predict_proba")
        if self.n_trt_ == 1:
            y = pred_diffs[0]
        else:
            y = np.dstack(pred_diffs)
        return y



"""Inverse propensity score-reweighted uplift model."""

from sklearn.utils.metaestimators import _BaseComposition
from sklearn.utils.metaestimators import available_if

class WrappedUpliftEstimator(_BaseComposition):
    """Wrap upift estimator to modify its functionality.

    If wrapped_predict is set it specifies the method to be called by
    wrapped model's predict.  This way standard scikit functions can
    access uplift specific predictions such as predict_action.
    """
    _uplift_model = True
    def __init__(self, base_estimator):
        self.base_estimator = base_estimator
    @property
    def _estimator_type(self):
        return self.base_estimator._estimator_type
    def fit(self, X, y, trt, n_trt, **kwargs):
        return self.base_estimator.fit(X, y, trt, n_trt, **kwargs)
    def score(self, X, y, trt, n_trt, **kwargs):
        return self.base_estimator.score(X, y, trt, n_trt, *args, **kwargs)
    @available_if(_estimator_has_predict())
    def predict(self, X):
        return self.base_estimator.predict(X)
    @available_if(_estimator_has("predict_action"))
    def predict_action(self, X):
        return self.base_estimator.predict_action(X)
    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X):
        return self.base_estimator.predict_proba(X)
    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X):
        return self.base_estimator.predict_log_proba(X)
    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        return self.base_estimator.decision_function(X)
    @available_if(_estimator_has("transform"))
    def transform(self, X):
        return self.base_estimator.transform(X)
    @available_if(_estimator_has("inverse_transform"))
    def inverse_transform(self, Xt):
        return self.base_estimator.inverse_transform(Xt)
    # generic parameters access
    def get_params(self, deep=True):
        if hasattr(self.base_estimator, "fit"):
            return super().get_params(deep=deep)
        return self._get_params('base_estimator', deep=deep)
    def set_params(self, **kwargs):
        if hasattr(self.base_estimator, "fit"):
            return super().set_params(**kwargs)
        else:
            self._set_params('base_estimator', **kwargs)
        return self
    def __getattr__(self, arg):
        return getattr(self.base_estimator, arg)


class IPTW_UpliftEstimator(WrappedUpliftEstimator):
    pass



"""Base classe for uplift meta models."""

import numpy as np

from sklearn.base import clone
from sklearn.utils import check_X_y, check_consistent_length
from sklearn.utils.metaestimators import _BaseComposition

from ..utils import check_trt

class UpliftMetaModelBase(_BaseComposition):
    """Base class for uplift meta estimators.

    Checks input consistency, builds classifiers on subsets of data.


    Derived classess need to overwride the _get_model_names_list and
    _iter_training_subsets methods.  The predict method needs to be
    implemented as well.

    """
    def __init__(self, base_estimator):
        self.base_estimator = base_estimator
    def _get_model_names_list(self, X=None, y=None, trt=None):
        """Return a list of names of constituent
        classification/regression models.

        This method should be overridden such that the number of
        models can be determined by _check_base_estimator.  If model
        name starts with '_' None will be put in model list instead of
        a real model and _ removed from name (useful to keep the list
        of given size even if some models are not used).
        """
        raise NotImplementedError()
    def _iter_training_subsets(self, X, y, trt, n_trt, sample_weight):
        """Return training sets for all models in the meta model.

        Each iteration returns a triple of predictor matrix, target
        vector, and sample weights (possibly None).  While iterating
        i-th subset it may be assumed that models on previous subsets
        have been fitted.

        """
        raise NotImplementedError()
    def _check_base_estimator(self, model_names):
        if hasattr(self.base_estimator, "fit"):
            estimator_list = []
            for m_name in model_names:
                if m_name.startswith("_"):
                    estimator_list.append((m_name[1:], None))
                else:
                    estimator_list.append((m_name, clone(self.base_estimator)))
        else:
            # full model list is provided, check length and names
            estimator_list = self.base_estimator
            new_estimator_list = []
            for m_name, (user_m_name, model) in zip(model_names, estimator_list):
                if m_name != user_m_name:
                    raise RuntimeError(f"Expected model name {m_name}, got {user_m_name}")
                if m_name.startswith("_") and model is not None:
                    raise RuntimeError(f"Model name {m_name} starts with '_' but the model is not None")
                if not m_name.startswith("_") and model is None:
                    raise RuntimeError(f"Model name {m_name} does not start with '_' but the model is None")
                new_estimator_list.append((m_name, clone(model)))
            estimator_list = new_estimator_list
        return estimator_list
    def fit(self, X, y, trt, n_trt=None, sample_weight=None):
        X, y = check_X_y(X, y, accept_sparse="csr")
        trt, n_trt = check_trt(trt, n_trt)
        check_consistent_length(X, y, trt)
        self._set_fit_params(y, trt, n_trt)

        self.models_ = self._check_base_estimator(self._get_model_names_list(X, y, trt))
        self.n_models_ = len(self.models_)
        self.n_ = np.zeros(self.n_models_, dtype=int)
        for i, (X_i, y_i, w_i) in enumerate(self._iter_training_subsets(X, y, trt, n_trt, sample_weight)):
            m_name, m_i = self.models_[i]
            if m_i is not None:
                if w_i is None:
                    self.n_[i] = X_i.shape[0]
                    m_i.fit(X_i, y_i)
                else:
                    self.n_[i] = w_i.sum()
                    m_i.fit(X_i, y_i, sample_weight=w_i)
        return self
    def get_params(self, deep=True):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        if hasattr(self.base_estimator, "fit"):
            return super().get_params(deep=deep)
        return self._get_params('base_estimator', deep=deep)
    def set_params(self, **kwargs):
        """Set the parameters of this estimator.
        Valid parameter keys can be listed with ``get_params()``.
        Returns
        -------
        self
        """
        if hasattr(self.base_estimator, "fit"):
            return super().set_params(**kwargs)
        else:
            self._set_params('base_estimator', **kwargs)
        return self
